import time
from typing import Dict, List, Optional, Union

from datahub.emitter.mcp_patch_builder import MetadataPatchProposal
from datahub.metadata.schema_classes import (
    AccessLevelClass,
    AuditStampClass,
    ChangeAuditStampsClass,
    DashboardInfoClass as DashboardInfo,
    EdgeClass as Edge,
    GlobalTagsClass as GlobalTags,
    GlossaryTermAssociationClass as Term,
    GlossaryTermsClass as GlossaryTerms,
    KafkaAuditHeaderClass,
    OwnerClass as Owner,
    OwnershipTypeClass,
    SystemMetadataClass,
    TagAssociationClass as Tag,
)
from datahub.specific.custom_properties import CustomPropertiesPatchHelper
from datahub.specific.ownership import OwnershipPatchHelper
from datahub.utilities.urns.tag_urn import TagUrn
from datahub.utilities.urns.urn import Urn


class DashboardPatchBuilder(MetadataPatchProposal):
    def __init__(
        self,
        urn: str,
        system_metadata: Optional[SystemMetadataClass] = None,
        audit_header: Optional[KafkaAuditHeaderClass] = None,
    ) -> None:
        """
        Initializes a DashboardPatchBuilder instance.

        Args:
            urn: The URN of the dashboard
            system_metadata: The system metadata of the dashboard (optional).
            audit_header: The Kafka audit header of the dashboard (optional).
        """
        super().__init__(
            urn, system_metadata=system_metadata, audit_header=audit_header
        )
        self.custom_properties_patch_helper = CustomPropertiesPatchHelper(
            self, DashboardInfo.ASPECT_NAME
        )
        self.ownership_patch_helper = OwnershipPatchHelper(self)

    def _mint_auditstamp(self, message: Optional[str] = None) -> AuditStampClass:
        """
        Creates an AuditStampClass instance with the current timestamp and other default values.

        Args:
            message: The message associated with the audit stamp (optional).

        Returns:
            An instance of AuditStampClass.
        """
        return AuditStampClass(
            time=int(time.time() * 1000.0),
            actor="urn:li:corpuser:datahub",
            message=message,
        )

    def _ensure_urn_type(
        self, entity_type: str, edges: List[Edge], context: str
    ) -> None:
        """
        Ensures that the destination URNs in the given edges have the specified entity type.

        Args:
            entity_type: The entity type to check against.
            edges: A list of Edge objects.
            context: The context or description of the operation.

        Raises:
            ValueError: If any of the destination URNs is not of the specified entity type.
        """
        for e in edges:
            urn = Urn.create_from_string(e.destinationUrn)
            if not urn.get_type() == entity_type:
                raise ValueError(
                    f"{context}: {e.destinationUrn} is not of type {entity_type}"
                )

    def add_owner(self, owner: Owner) -> "DashboardPatchBuilder":
        """
        Adds an owner to the DashboardPatchBuilder.

        Args:
            owner: The Owner object to add.

        Returns:
            The DashboardPatchBuilder instance.
        """
        self.ownership_patch_helper.add_owner(owner)
        return self

    def remove_owner(
        self, owner: str, owner_type: Optional[OwnershipTypeClass] = None
    ) -> "DashboardPatchBuilder":
        """
        Removes an owner from the DashboardPatchBuilder.

        Args:
            owner: The owner to remove.
            owner_type: The ownership type of the owner (optional).

        Returns:
            The DashboardPatchBuilder instance.

        Notes:
            `owner_type` is optional.
        """
        self.ownership_patch_helper.remove_owner(owner, owner_type)
        return self

    def set_owners(self, owners: List[Owner]) -> "DashboardPatchBuilder":
        """
        Sets the owners of the DashboardPatchBuilder.

        Args:
            owners: A list of Owner objects.

        Returns:
            The DashboardPatchBuilder instance.
        """
        self.ownership_patch_helper.set_owners(owners)
        return self

    def add_dataset_edge(
        self, dataset: Union[Edge, Urn, str]
    ) -> "DashboardPatchBuilder":
        """
        Adds an dataset to the DashboardPatchBuilder.

        Args:
            dataset: The dataset, which can be an Edge object, Urn object, or a string.

        Returns:
            The DashboardPatchBuilder instance.

        Raises:
            ValueError: If the dataset is not a Dataset urn.

        Notes:
            If `dataset` is an Edge object, it is used directly. If `dataset` is a Urn object or string,
            it is converted to an Edge object and added with default audit stamps.
        """
        if isinstance(dataset, Edge):
            dataset_urn: str = dataset.destinationUrn
            dataset_edge: Edge = dataset
        elif isinstance(dataset, (Urn, str)):
            dataset_urn = str(dataset)
            if not dataset_urn.startswith("urn:li:dataset:"):
                raise ValueError(f"Input {dataset} is not a Dataset urn")

            dataset_edge = Edge(
                destinationUrn=dataset_urn,
                created=self._mint_auditstamp(),
                lastModified=self._mint_auditstamp(),
            )

        self._ensure_urn_type("dataset", [dataset_edge], "add_dataset")
        self._add_patch(
            DashboardInfo.ASPECT_NAME,
            "add",
            path=f"/datasetEdges/{self.quote(dataset_urn)}",
            value=dataset_edge,
        )
        return self

    def remove_dataset_edge(self, dataset: Union[str, Urn]) -> "DashboardPatchBuilder":
        """
        Removes a dataset edge from the DashboardPatchBuilder.

        Args:
            dataset: The dataset to remove, specified as a string or Urn object.

        Returns:
            The DashboardPatchBuilder instance.
        """
        self._add_patch(
            DashboardInfo.ASPECT_NAME,
            "remove",
            path=f"/datasetEdges/{dataset}",
            value={},
        )
        return self

    def set_dataset_edges(self, datasets: List[Edge]) -> "DashboardPatchBuilder":
        """
        Sets the dataset edges for the DashboardPatchBuilder.

        Args:
            datasets: A list of Edge objects representing the dataset edges.

        Returns:
            The DashboardPatchBuilder instance.

        Raises:
            ValueError: If any of the input edges are not of type 'Datset'.

        Notes:
            This method replaces all existing datasets with the given inputs.
        """
        self._ensure_urn_type("dataset", datasets, "dataset edges")
        self._add_patch(
            DashboardInfo.ASPECT_NAME,
            "add",
            path="/datasetEdges",
            value=datasets,
        )
        return self

    def add_chart_edge(self, chart: Union[Edge, Urn, str]) -> "DashboardPatchBuilder":
        """
        Adds a chart edge to the DashboardPatchBuilder.

        Args:
            chart: The dataset, which can be an Edge object, Urn object, or a string.

        Returns:
            The DashboardPatchBuilder instance.

        Raises:
            ValueError: If the edge is not a Chart urn.

        Notes:
            If `chart` is an Edge object, it is used directly. If `chart` is a Urn object or string,
            it is converted to an Edge object and added with default audit stamps.
        """
        if isinstance(chart, Edge):
            chart_urn: str = chart.destinationUrn
            chart_edge: Edge = chart
        elif isinstance(chart, (Urn, str)):
            chart_urn = str(chart)
            if not chart_urn.startswith("urn:li:chart:"):
                raise ValueError(f"Input {chart} is not a Chart urn")

            chart_edge = Edge(
                destinationUrn=chart_urn,
                created=self._mint_auditstamp(),
                lastModified=self._mint_auditstamp(),
            )

        self._ensure_urn_type("dataset", [chart_edge], "add_chart_edge")
        self._add_patch(
            DashboardInfo.ASPECT_NAME,
            "add",
            path=f"/chartEdges/{self.quote(chart_urn)}",
            value=chart_edge,
        )
        return self

    def remove_chart_edge(self, chart: Union[str, Urn]) -> "DashboardPatchBuilder":
        """
        Removes an chart edge from the DashboardPatchBuilder.

        Args:
            chart: The chart to remove, specified as a string or Urn object.

        Returns:
            The DashboardPatchBuilder instance.
        """
        self._add_patch(
            DashboardInfo.ASPECT_NAME,
            "remove",
            path=f"/chartEdges/{chart}",
            value={},
        )
        return self

    def set_chart_edges(self, charts: List[Edge]) -> "DashboardPatchBuilder":
        """
        Sets the chart edges for the DashboardPatchBuilder.

        Args:
            charts: A list of Edge objects representing the chart edges.

        Returns:
            The DashboardPatchBuilder instance.

        Raises:
            ValueError: If any of the edges are not of type 'chart'.

        Notes:
            This method replaces all existing charts with the given charts.
        """
        self._ensure_urn_type("chart", charts, "set_charts")
        self._add_patch(
            DashboardInfo.ASPECT_NAME,
            "add",
            path="/chartEdges",
            value=charts,
        )
        return self

    def add_tag(self, tag: Tag) -> "DashboardPatchBuilder":
        """
        Adds a tag to the DashboardPatchBuilder.

        Args:
            tag: The Tag object representing the tag to be added.

        Returns:
            The DashboardPatchBuilder instance.
        """
        self._add_patch(
            GlobalTags.ASPECT_NAME, "add", path=f"/tags/{tag.tag}", value=tag
        )
        return self

    def remove_tag(self, tag: Union[str, Urn]) -> "DashboardPatchBuilder":
        """
        Removes a tag from the DashboardPatchBuilder.

        Args:
            tag: The tag to remove, specified as a string or Urn object.

        Returns:
            The DashboardPatchBuilder instance.
        """
        if isinstance(tag, str) and not tag.startswith("urn:li:tag:"):
            tag = TagUrn.create_from_id(tag)
        self._add_patch(GlobalTags.ASPECT_NAME, "remove", path=f"/tags/{tag}", value={})
        return self

    def add_term(self, term: Term) -> "DashboardPatchBuilder":
        """
        Adds a glossary term to the DashboardPatchBuilder.

        Args:
            term: The Term object representing the glossary term to be added.

        Returns:
            The DashboardPatchBuilder instance.
        """
        self._add_patch(
            GlossaryTerms.ASPECT_NAME, "add", path=f"/terms/{term.urn}", value=term
        )
        return self

    def remove_term(self, term: Union[str, Urn]) -> "DashboardPatchBuilder":
        """
        Removes a glossary term from the DashboardPatchBuilder.

        Args:
            term: The term to remove, specified as a string or Urn object.

        Returns:
            The DashboardPatchBuilder instance.
        """
        if isinstance(term, str) and not term.startswith("urn:li:glossaryTerm:"):
            term = "urn:li:glossaryTerm:" + term
        self._add_patch(
            GlossaryTerms.ASPECT_NAME, "remove", path=f"/terms/{term}", value={}
        )
        return self

    def set_custom_properties(
        self, custom_properties: Dict[str, str]
    ) -> "DashboardPatchBuilder":
        """
        Sets the custom properties for the DashboardPatchBuilder.

        Args:
            custom_properties: A dictionary containing the custom properties to be set.

        Returns:
            The DashboardPatchBuilder instance.

        Notes:
            This method replaces all existing custom properties with the given dictionary.
        """
        self._add_patch(
            DashboardInfo.ASPECT_NAME,
            "add",
            path="/customProperties",
            value=custom_properties,
        )
        return self

    def add_custom_property(self, key: str, value: str) -> "DashboardPatchBuilder":
        """
        Adds a custom property to the DashboardPatchBuilder.

        Args:
            key: The key of the custom property.
            value: The value of the custom property.

        Returns:
            The DashboardPatchBuilder instance.
        """
        self.custom_properties_patch_helper.add_property(key, value)
        return self

    def remove_custom_property(self, key: str) -> "DashboardPatchBuilder":
        """
        Removes a custom property from the DashboardPatchBuilder.

        Args:
            key: The key of the custom property to remove.

        Returns:
            The DashboardPatchBuilder instance.
        """
        self.custom_properties_patch_helper.remove_property(key)
        return self

    def set_title(self, title: str) -> "DashboardPatchBuilder":
        assert title, "DashboardInfo title should not be None"
        self._add_patch(
            DashboardInfo.ASPECT_NAME,
            "add",
            path="/title",
            value=title,
        )

        return self

    def set_description(self, description: str) -> "DashboardPatchBuilder":
        assert description, "DashboardInfo description should not be None"
        self._add_patch(
            DashboardInfo.ASPECT_NAME,
            "add",
            path="/description",
            value=description,
        )

        return self

    def add_custom_properties(
        self, custom_properties: Optional[Dict[str, str]] = None
    ) -> "DashboardPatchBuilder":

        if custom_properties:
            for key, value in custom_properties.items():
                self.custom_properties_patch_helper.add_property(key, value)

        return self

    def set_external_url(self, external_url: Optional[str]) -> "DashboardPatchBuilder":
        if external_url:
            self._add_patch(
                DashboardInfo.ASPECT_NAME,
                "add",
                path="/externalUrl",
                value=external_url,
            )
        return self

    def add_charts(self, chart_urns: Optional[List[str]]) -> "DashboardPatchBuilder":
        if chart_urns:
            for urn in chart_urns:
                self._add_patch(
                    aspect_name=DashboardInfo.ASPECT_NAME,
                    op="add",
                    path=f"/charts/{urn}",
                    value=urn,
                )

        return self

    def add_datasets(
        self, dataset_urns: Optional[List[str]]
    ) -> "DashboardPatchBuilder":
        if dataset_urns:
            for urn in dataset_urns:
                self._add_patch(
                    aspect_name=DashboardInfo.ASPECT_NAME,
                    op="add",
                    path=f"/datasets/{urn}",
                    value=urn,
                )

        return self

    def set_dashboard_url(
        self, dashboard_url: Optional[str]
    ) -> "DashboardPatchBuilder":
        if dashboard_url:
            self._add_patch(
                DashboardInfo.ASPECT_NAME,
                "add",
                path="/dashboardUrl",
                value=dashboard_url,
            )

        return self

    def set_access(
        self, access: Union[None, Union[str, "AccessLevelClass"]] = None
    ) -> "DashboardPatchBuilder":
        if access:
            self._add_patch(
                DashboardInfo.ASPECT_NAME,
                "add",
                path="/access",
                value=access,
            )

        return self

    def set_last_refreshed(
        self, last_refreshed: Optional[int]
    ) -> "DashboardPatchBuilder":
        if last_refreshed:
            self._add_patch(
                DashboardInfo.ASPECT_NAME,
                "add",
                path="/lastRefreshed",
                value=last_refreshed,
            )

        return self

    def set_last_modified(
        self, last_modified: "ChangeAuditStampsClass"
    ) -> "DashboardPatchBuilder":
        if last_modified:
            self._add_patch(
                DashboardInfo.ASPECT_NAME,
                "add",
                path="/lastModified",
                value=last_modified,
            )

        return self



import json
import logging
import ssl
import time
from dataclasses import dataclass, field
from datetime import datetime, timedelta, timezone
from enum import Enum
from typing import Callable, Dict, Iterable, List, Optional, Tuple, Union
from urllib.parse import urljoin

import requests
from cached_property import cached_property
from dateutil import parser
from packaging import version
from pydantic import root_validator, validator
from pydantic.fields import Field
from requests import Response
from requests.adapters import HTTPAdapter
from requests.models import HTTPBasicAuth
from requests_gssapi import HTTPSPNEGOAuth

import datahub.emitter.mce_builder as builder
from datahub.configuration.common import AllowDenyPattern
from datahub.configuration.source_common import EnvConfigMixin
from datahub.emitter.mcp import MetadataChangeProposalWrapper
from datahub.emitter.mcp_builder import ContainerKey, gen_containers
from datahub.ingestion.api.common import PipelineContext
from datahub.ingestion.api.decorators import (
    SupportStatus,
    capability,
    config_class,
    platform_name,
    support_status,
)
from datahub.ingestion.api.source import Source, SourceCapability, SourceReport
from datahub.ingestion.api.workunit import MetadataWorkUnit
from datahub.ingestion.source.common.subtypes import JobContainerSubTypes
from datahub.metadata.schema_classes import (
    BrowsePathEntryClass,
    BrowsePathsV2Class,
    DataFlowInfoClass,
    DataJobInfoClass,
    DataJobInputOutputClass,
    DataPlatformInstanceClass,
    DatasetPropertiesClass,
)
from datahub.specific.datajob import DataJobPatchBuilder

logger = logging.getLogger(__name__)
NIFI = "nifi"


# Python requests does not support passing password for key file,
# The same can be achieved by mounting ssl context
# as described here - https://github.com/psf/requests/issues/2519
# and here - https://github.com/psf/requests/issues/1573
class SSLAdapter(HTTPAdapter):
    def __init__(self, certfile, keyfile, password=None):
        self.context = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)
        self.context.load_cert_chain(
            certfile=certfile, keyfile=keyfile, password=password
        )
        super().__init__()

    def init_poolmanager(self, *args, **kwargs):
        kwargs["ssl_context"] = self.context
        return super().init_poolmanager(*args, **kwargs)


class NifiAuthType(Enum):
    NO_AUTH = "NO_AUTH"
    SINGLE_USER = "SINGLE_USER"
    CLIENT_CERT = "CLIENT_CERT"
    KERBEROS = "KERBEROS"
    BASIC_AUTH = "BASIC_AUTH"


class ProcessGroupKey(ContainerKey):
    process_group_id: str


class NifiSourceConfig(EnvConfigMixin):
    site_url: str = Field(
        description="URL for Nifi, ending with /nifi/. e.g. https://mynifi.domain/nifi/"
    )

    auth: NifiAuthType = Field(
        default=NifiAuthType.NO_AUTH,
        description="Nifi authentication. must be one of : NO_AUTH, SINGLE_USER, CLIENT_CERT, KERBEROS",
    )

    provenance_days: int = Field(
        default=7,
        description="time window to analyze provenance events for external datasets",
    )  # Fetch provenance events for past 1 week
    process_group_pattern: AllowDenyPattern = Field(
        default=AllowDenyPattern.allow_all(),
        description="regex patterns for filtering process groups",
    )

    # Required for nifi deployments using Remote Process Groups
    site_name: str = Field(
        default="default",
        description="Site name to identify this site with, useful when using input and output ports receiving remote connections",
    )
    site_url_to_site_name: Dict[str, str] = Field(
        default={},
        description="Lookup to find site_name for site_url ending with /nifi/, required if using remote process groups in nifi flow",
    )

    # Required to be set if auth is of type SINGLE_USER
    username: Optional[str] = Field(
        default=None, description='Nifi username, must be set for auth = "SINGLE_USER"'
    )
    password: Optional[str] = Field(
        default=None, description='Nifi password, must be set for auth = "SINGLE_USER"'
    )

    # Required to be set if auth is of type CLIENT_CERT
    client_cert_file: Optional[str] = Field(
        default=None,
        description='Path to PEM file containing the public certificates for the user/client identity, must be set for auth = "CLIENT_CERT"',
    )
    client_key_file: Optional[str] = Field(
        default=None, description="Path to PEM file containing the client’s secret key"
    )
    client_key_password: Optional[str] = Field(
        default=None, description="The password to decrypt the client_key_file"
    )

    # Required to be set if nifi server certificate is not signed by
    # root CA trusted by client system, e.g. self-signed certificates
    ca_file: Optional[Union[bool, str]] = Field(
        default=None,
        description="Path to PEM file containing certs for the root CA(s) for the NiFi."
        "Set to False to disable SSL verification.",
    )

    # As of now, container entities retrieval does not respect browsePathsV2 similar to container aspect.
    # Consider enabling this when entities with browsePathsV2 pointing to container also get listed in container entities.
    emit_process_group_as_container: bool = Field(
        default=False,
        description="Whether to emit Nifi process groups as container entities.",
    )

    incremental_lineage: bool = Field(
        default=True,
        description="When enabled, emits incremental/patch lineage for Nifi processors."
        " When disabled, re-states lineage on each run.",
    )

    @root_validator(skip_on_failure=True)
    def validate_auth_params(cla, values):
        if values.get("auth") is NifiAuthType.CLIENT_CERT and not values.get(
            "client_cert_file"
        ):
            raise ValueError(
                "Config `client_cert_file` is required for CLIENT_CERT auth"
            )
        elif values.get("auth") in (
            NifiAuthType.SINGLE_USER,
            NifiAuthType.BASIC_AUTH,
        ) and (not values.get("username") or not values.get("password")):
            raise ValueError(
                f"Config `username` and `password` is required for {values.get('auth').value} auth"
            )
        return values

    @root_validator(skip_on_failure=True)
    def validator_site_url_to_site_name(cls, values):
        site_url_to_site_name = values.get("site_url_to_site_name")
        site_url = values.get("site_url")
        site_name = values.get("site_name")

        if site_url_to_site_name is None:
            site_url_to_site_name = {}
            values["site_url_to_site_name"] = site_url_to_site_name

        if site_url not in site_url_to_site_name:
            site_url_to_site_name[site_url] = site_name

        return values

    @validator("site_url")
    def validator_site_url(cls, site_url: str) -> str:
        assert site_url.startswith(
            ("http://", "https://")
        ), "site_url must start with http:// or https://"

        if not site_url.endswith("/"):
            site_url = site_url + "/"

        if not site_url.endswith("/nifi/"):
            site_url = site_url + "nifi/"

        return site_url


TOKEN_ENDPOINT = "access/token"
KERBEROS_TOKEN_ENDPOINT = "access/kerberos"
ABOUT_ENDPOINT = "flow/about"
CLUSTER_ENDPOINT = "flow/cluster/summary"
PG_ENDPOINT = "flow/process-groups/"
PROVENANCE_ENDPOINT = "provenance"


class NifiType(Enum):
    PROCESSOR = "PROCESSOR"
    FUNNEL = "FUNNEL"
    INPUT_PORT = "INPUT_PORT"
    OUTPUT_PORT = "OUTPUT_PORT"
    REMOTE_INPUT_PORT = "REMOTE_INPUT_PORT"
    REMOTE_OUTPUT_PORT = "REMOTE_OUTPUT_PORT"


class NifiEventType:
    CREATE = "CREATE"
    FETCH = "FETCH"
    SEND = "SEND"
    RECEIVE = "RECEIVE"


class NifiProcessorType:
    ListS3 = "org.apache.nifi.processors.aws.s3.ListS3"
    FetchS3Object = "org.apache.nifi.processors.aws.s3.FetchS3Object"
    PutS3Object = "org.apache.nifi.processors.aws.s3.PutS3Object"
    ListSFTP = "org.apache.nifi.processors.standard.ListSFTP"
    FetchSFTP = "org.apache.nifi.processors.standard.FetchSFTP"
    GetSFTP = "org.apache.nifi.processors.standard.GetSFTP"
    PutSFTP = "org.apache.nifi.processors.standard.PutSFTP"


# To support new processor type,
# 1. add an entry in KNOWN_INGRESS_EGRESS_PROCESORS
# 2. Implement provenance event analyzer to find external dataset and
# map it in provenance_event_to_lineage_map
class NifiProcessorProvenanceEventAnalyzer:
    env: str

    KNOWN_INGRESS_EGRESS_PROCESORS = {
        NifiProcessorType.ListS3: NifiEventType.CREATE,
        NifiProcessorType.FetchS3Object: NifiEventType.FETCH,
        NifiProcessorType.PutS3Object: NifiEventType.SEND,
        NifiProcessorType.ListSFTP: NifiEventType.CREATE,
        NifiProcessorType.FetchSFTP: NifiEventType.FETCH,
        NifiProcessorType.GetSFTP: NifiEventType.RECEIVE,
        NifiProcessorType.PutSFTP: NifiEventType.SEND,
    }

    def __init__(self) -> None:
        # Map of Nifi processor type to the provenance event analyzer to find lineage
        self.provenance_event_to_lineage_map: Dict[
            str, Callable[[Dict], ExternalDataset]
        ] = {
            NifiProcessorType.ListS3: self.process_s3_provenance_event,
            NifiProcessorType.FetchS3Object: self.process_s3_provenance_event,
            NifiProcessorType.PutS3Object: self.process_s3_provenance_event,
            NifiProcessorType.ListSFTP: self.process_sftp_provenance_event,
            NifiProcessorType.FetchSFTP: self.process_sftp_provenance_event,
            NifiProcessorType.GetSFTP: self.process_sftp_provenance_event,
            NifiProcessorType.PutSFTP: self.process_sftp_provenance_event,
        }

    def process_s3_provenance_event(self, event):
        attributes = event.get("attributes", [])
        s3_bucket = get_attribute_value(attributes, "s3.bucket")
        s3_key = get_attribute_value(attributes, "s3.key")
        if not s3_key:
            s3_key = get_attribute_value(attributes, "filename")

        s3_url = f"s3://{s3_bucket}/{s3_key}"
        s3_url = s3_url[: s3_url.rindex("/")]
        s3_path = s3_url[len("s3://") :]
        dataset_name = s3_path.replace("/", ".")
        platform = "s3"
        dataset_urn = builder.make_dataset_urn(platform, s3_path, self.env)
        return ExternalDataset(
            platform,
            dataset_name,
            dict(s3_uri=s3_url),
            dataset_urn,
        )

    def process_sftp_provenance_event(self, event):
        attributes = event.get("attributes", [])
        remote_host = get_attribute_value(attributes, "sftp.remote.host")
        path = get_attribute_value(attributes, "path")
        filename = get_attribute_value(attributes, "filename")
        absolute_path = f"sftp://{remote_host}/{path}/{filename}"
        if remote_host is None or path is None or filename is None:
            absolute_path = event.get("transitUri")

        absolute_path = absolute_path.replace("/./", "/")
        if absolute_path.endswith("/."):
            absolute_path = absolute_path[:-2]
        absolute_path = absolute_path[: absolute_path.rindex("/")]
        dataset_name = absolute_path.replace("sftp://", "").replace("/", ".")
        platform = "file"
        dataset_urn = builder.make_dataset_urn(platform, dataset_name, self.env)
        return ExternalDataset(
            platform,
            dataset_name,
            dict(uri=absolute_path),
            dataset_urn,
        )


@dataclass
class ExternalDataset:
    platform: str
    dataset_name: str
    dataset_properties: Dict[str, str]
    dataset_urn: str


@dataclass
class NifiComponent:
    id: str
    name: str
    type: str
    parent_group_id: str
    nifi_type: NifiType
    comments: Optional[str] = None
    status: Optional[str] = None

    # present only for nifi remote ports and processors
    inlets: Dict[str, ExternalDataset] = field(default_factory=dict)
    outlets: Dict[str, ExternalDataset] = field(default_factory=dict)

    # present only for processors
    config: Optional[Dict] = None

    # present only for nifi remote ports
    target_uris: Optional[str] = None
    parent_rpg_id: Optional[str] = None

    # Last successful event time
    last_event_time: Optional[str] = None


@dataclass
class NifiProcessGroup:
    id: str
    name: str
    parent_group_id: Optional[str]


@dataclass
class NifiRemoteProcessGroup:
    id: str
    name: str
    parent_group_id: str
    remote_ports: Dict[str, NifiComponent]


@dataclass
class NifiFlow:
    version: Optional[str]
    clustered: Optional[bool]
    root_process_group: NifiProcessGroup
    components: Dict[str, NifiComponent] = field(default_factory=dict)
    remotely_accessible_ports: Dict[str, NifiComponent] = field(default_factory=dict)
    connections: List[Tuple[str, str]] = field(default_factory=list)
    processGroups: Dict[str, NifiProcessGroup] = field(default_factory=dict)
    remoteProcessGroups: Dict[str, NifiRemoteProcessGroup] = field(default_factory=dict)
    remote_ports: Dict[str, NifiComponent] = field(default_factory=dict)


def get_attribute_value(attr_lst: List[dict], attr_name: str) -> Optional[str]:
    match = [entry for entry in attr_lst if entry["name"] == attr_name]
    if len(match) > 0:
        return match[0]["value"]
    return None


@dataclass
class NifiSourceReport(SourceReport):
    filtered: List[str] = field(default_factory=list)

    def report_dropped(self, ent_name: str) -> None:
        self.filtered.append(ent_name)


# allowRemoteAccess
@platform_name("NiFi", id="nifi")
@config_class(NifiSourceConfig)
@support_status(SupportStatus.CERTIFIED)
@capability(SourceCapability.LINEAGE_COARSE, "Supported. See docs for limitations")
class NifiSource(Source):

    config: NifiSourceConfig
    report: NifiSourceReport

    def __init__(self, config: NifiSourceConfig, ctx: PipelineContext) -> None:
        super().__init__(ctx)
        self.config = config
        self.report = NifiSourceReport()
        self.session = requests.Session()

        if self.config.ca_file is not None:
            self.session.verify = self.config.ca_file

        # To keep track of process groups (containers) which have already been ingested
        # Required, as we do not ingest all process groups but only those that have known ingress/egress processors
        self.processed_pgs: List[str] = []

    @cached_property
    def rest_api_base_url(self):
        return self.config.site_url[: -len("nifi/")] + "nifi-api/"

    @classmethod
    def create(cls, config_dict: dict, ctx: PipelineContext) -> "Source":
        config = NifiSourceConfig.parse_obj(config_dict)
        return cls(config, ctx)

    def get_report(self) -> SourceReport:
        return self.report

    def update_flow(self, pg_flow_dto: Dict) -> None:  # noqa: C901
        """
        Update self.nifi_flow with contents of the input process group `pg_flow_dto`
        """
        breadcrumb_dto = pg_flow_dto.get("breadcrumb", {}).get("breadcrumb", {})
        nifi_pg = NifiProcessGroup(
            breadcrumb_dto.get("id"),
            breadcrumb_dto.get("name"),
            pg_flow_dto.get("parentGroupId"),
        )
        self.nifi_flow.processGroups[nifi_pg.id] = nifi_pg
        if not self.config.process_group_pattern.allowed(nifi_pg.name):
            self.report.report_dropped(f"{nifi_pg.name}.*")
            return

        flow_dto = pg_flow_dto.get("flow", {})

        for processor in flow_dto.get("processors", []):
            component = processor.get("component")
            self.nifi_flow.components[component.get("id")] = NifiComponent(
                component.get("id"),
                component.get("name"),
                component.get("type"),
                component.get("parentGroupId"),
                NifiType.PROCESSOR,
                config=component.get("config"),
                comments=component.get("config", {}).get("comments"),
                status=component.get("status", {}).get("runStatus"),
            )
        for funnel in flow_dto.get("funnels", []):
            component = funnel.get("component")
            self.nifi_flow.components[component.get("id")] = NifiComponent(
                component.get("id"),
                component.get("name"),
                component.get("type"),
                component.get("parentGroupId"),
                NifiType.FUNNEL,
                comments=component.get("comments"),
                status=component.get("status", {}).get("runStatus"),
            )
            logger.debug(f"Adding funnel {component.get('id')}")

        for connection in flow_dto.get("connections", []):
            # Exclude self - recursive relationships
            if connection.get("sourceId") != connection.get("destinationId"):
                self.nifi_flow.connections.append(
                    (connection.get("sourceId"), connection.get("destinationId"))
                )

        for inputPort in flow_dto.get("inputPorts", []):
            component = inputPort.get("component")
            if inputPort.get("allowRemoteAccess"):
                self.nifi_flow.remotely_accessible_ports[
                    component.get("id")
                ] = NifiComponent(
                    component.get("id"),
                    component.get("name"),
                    component.get("type"),
                    component.get("parentGroupId"),
                    NifiType.INPUT_PORT,
                    comments=component.get("comments"),
                    status=component.get("status", {}).get("runStatus"),
                )
                logger.debug(f"Adding remotely accessible port {component.get('id')}")
            else:
                self.nifi_flow.components[component.get("id")] = NifiComponent(
                    component.get("id"),
                    component.get("name"),
                    component.get("type"),
                    component.get("parentGroupId"),
                    NifiType.INPUT_PORT,
                    comments=component.get("comments"),
                    status=component.get("status", {}).get("runStatus"),
                )
                logger.debug(f"Adding port {component.get('id')}")

        for outputPort in flow_dto.get("outputPorts", []):
            component = outputPort.get("component")
            if outputPort.get("allowRemoteAccess"):
                self.nifi_flow.remotely_accessible_ports[
                    component.get("id")
                ] = NifiComponent(
                    component.get("id"),
                    component.get("name"),
                    component.get("type"),
                    component.get("parentGroupId"),
                    NifiType.OUTPUT_PORT,
                    comments=component.get("comments"),
                    status=component.get("status", {}).get("runStatus"),
                )
                logger.debug(f"Adding remotely accessible port {component.get('id')}")
            else:
                self.nifi_flow.components[component.get("id")] = NifiComponent(
                    component.get("id"),
                    component.get("name"),
                    component.get("type"),
                    component.get("parentGroupId"),
                    NifiType.OUTPUT_PORT,
                    comments=component.get("comments"),
                    status=component.get("status", {}).get("runStatus"),
                )
                logger.debug(f"Adding report port {component.get('id')}")

        for rpg in flow_dto.get("remoteProcessGroups", []):
            rpg_component = rpg.get("component", {})
            remote_ports = {}

            contents = rpg_component.get("contents", {})
            for component in contents.get("outputPorts", []):
                if component.get("connected", False):
                    remote_ports[component.get("id")] = NifiComponent(
                        component.get("id"),
                        component.get("name"),
                        component.get("type"),
                        rpg_component.get("parentGroupId"),
                        NifiType.REMOTE_OUTPUT_PORT,
                        target_uris=rpg_component.get("targetUris"),
                        parent_rpg_id=rpg_component.get("id"),
                        comments=component.get("comments"),
                        status=component.get("status", {}).get("runStatus"),
                    )
                    logger.debug(f"Adding remote output port {component.get('id')}")

            for component in contents.get("inputPorts", []):
                if component.get("connected", False):
                    remote_ports[component.get("id")] = NifiComponent(
                        component.get("id"),
                        component.get("name"),
                        component.get("type"),
                        rpg_component.get("parentGroupId"),
                        NifiType.REMOTE_INPUT_PORT,
                        target_uris=rpg_component.get("targetUris"),
                        parent_rpg_id=rpg_component.get("id"),
                        comments=component.get("comments"),
                        status=component.get("status", {}).get("runStatus"),
                    )
                    logger.debug(f"Adding remote input port {component.get('id')}")

            nifi_rpg = NifiRemoteProcessGroup(
                rpg_component.get("id"),
                rpg_component.get("name"),
                component.get("parentGroupId"),
                remote_ports,
            )
            logger.debug(f"Adding remote process group {rpg_component.get('id')}")
            self.nifi_flow.components.update(remote_ports)
            self.nifi_flow.remoteProcessGroups[nifi_rpg.id] = nifi_rpg

        for pg in flow_dto.get("processGroups", []):
            pg_response = self.session.get(
                url=urljoin(self.rest_api_base_url, PG_ENDPOINT) + pg.get("id")
            )

            if not pg_response.ok:
                self.report.warning(
                    "Failed to get process group flow " + pg.get("id"),
                    self.config.site_url,
                )
                continue

            pg_flow_dto = pg_response.json().get("processGroupFlow", {})

            self.update_flow(pg_flow_dto)

    def update_flow_keep_only_ingress_egress(self):
        components_to_del: List[NifiComponent] = []
        for component in self.nifi_flow.components.values():
            if (
                component.nifi_type is NifiType.PROCESSOR
                and component.type
                not in NifiProcessorProvenanceEventAnalyzer.KNOWN_INGRESS_EGRESS_PROCESORS.keys()
            ) or component.nifi_type not in [
                NifiType.PROCESSOR,
                NifiType.REMOTE_INPUT_PORT,
                NifiType.REMOTE_OUTPUT_PORT,
            ]:
                components_to_del.append(component)
                incoming = list(
                    filter(lambda x: x[1] == component.id, self.nifi_flow.connections)
                )
                outgoing = list(
                    filter(lambda x: x[0] == component.id, self.nifi_flow.connections)
                )
                # Create new connections from incoming to outgoing
                for i in incoming:
                    for j in outgoing:
                        self.nifi_flow.connections.append((i[0], j[1]))

                # Remove older connections, as we already created
                # new connections bypassing component to be deleted

                for i in incoming:
                    self.nifi_flow.connections.remove(i)
                for j in outgoing:
                    self.nifi_flow.connections.remove(j)

        for c in components_to_del:
            if c.nifi_type is NifiType.PROCESSOR and (
                c.name.startswith("Get")
                or c.name.startswith("List")
                or c.name.startswith("Fetch")
                or c.name.startswith("Put")
            ):
                self.report.warning(
                    f"Dropping NiFi Processor of type {c.type}, id {c.id}, name {c.name} from lineage view. \
                    This is likely an Ingress or Egress node which may be reading to/writing from external datasets \
                    However not currently supported in datahub",
                    self.config.site_url,
                )
            else:
                logger.debug(
                    f"Dropping NiFi Component of type {c.type}, id {c.id}, name {c.name} from lineage view."
                )

            del self.nifi_flow.components[c.id]

    def create_nifi_flow(self):
        about_response = self.session.get(
            url=urljoin(self.rest_api_base_url, ABOUT_ENDPOINT)
        )
        nifi_version: Optional[str] = None
        if about_response.ok:
            try:
                nifi_version = about_response.json().get("about", {}).get("version")
            except Exception as e:
                logger.error(
                    f"Unable to parse about response from Nifi: {about_response} due to {e}"
                )
        else:
            logger.warning("Failed to fetch version for nifi")
        cluster_response = self.session.get(
            url=urljoin(self.rest_api_base_url, CLUSTER_ENDPOINT)
        )
        clustered: Optional[bool] = None
        if cluster_response.ok:
            clustered = (
                cluster_response.json().get("clusterSummary", {}).get("clustered")
            )
        else:
            logger.warning("Failed to fetch cluster summary for flow")
        pg_response = self.session.get(
            url=urljoin(self.rest_api_base_url, PG_ENDPOINT) + "root"
        )

        pg_response.raise_for_status()

        pg_flow_dto = pg_response.json().get("processGroupFlow", {})
        breadcrumb_dto = pg_flow_dto.get("breadcrumb", {}).get("breadcrumb", {})
        self.nifi_flow = NifiFlow(
            version=nifi_version,
            clustered=clustered,
            root_process_group=NifiProcessGroup(
                breadcrumb_dto.get("id"),
                breadcrumb_dto.get("name"),
                pg_flow_dto.get("parentGroupId"),
            ),
        )
        self.update_flow(pg_flow_dto)
        self.update_flow_keep_only_ingress_egress()

    def fetch_provenance_events(
        self,
        processor: NifiComponent,
        eventType: str,
        startDate: datetime,
        endDate: Optional[datetime] = None,
    ) -> Iterable[Dict]:
        logger.debug(
            f"Fetching {eventType} provenance events for {processor.id}\
            of processor type {processor.type}, Start date: {startDate}, End date: {endDate}"
        )

        provenance_response = self.submit_provenance_query(
            processor, eventType, startDate, endDate
        )

        if provenance_response.ok:
            provenance = provenance_response.json().get("provenance", {})
            provenance_uri = provenance.get("uri")

            provenance_response = self.session.get(provenance_uri)
            if provenance_response.ok:
                provenance = provenance_response.json().get("provenance", {})

            attempts = 5  # wait for at most 5 attempts 5*1= 5 seconds
            while (not provenance.get("finished", False)) and attempts > 0:
                logger.warning(
                    f"Provenance query not completed, attempts left : {attempts}"
                )
                # wait until the uri returns percentcomplete 100
                time.sleep(1)
                provenance_response = self.session.get(provenance_uri)
                attempts -= 1
                if provenance_response.ok:
                    provenance = provenance_response.json().get("provenance", {})

            events = provenance.get("results", {}).get("provenanceEvents", [])
            last_event_time: Optional[datetime] = None
            oldest_event_time: Optional[datetime] = None

            for event in events:
                event_time = parser.parse(event.get("eventTime"))
                # datetime.strptime(
                #    event.get("eventTime"), "%m/%d/%Y %H:%M:%S.%f %Z"
                # )
                if not last_event_time or event_time > last_event_time:
                    last_event_time = event_time

                if not oldest_event_time or event_time < oldest_event_time:
                    oldest_event_time = event_time

                yield event

            processor.last_event_time = str(last_event_time)
            self.delete_provenance(provenance_uri)

            total = provenance.get("results", {}).get("total")
            totalCount = provenance.get("results", {}).get("totalCount")
            if total != str(totalCount):
                yield from self.fetch_provenance_events(
                    processor, eventType, startDate, oldest_event_time
                )
        else:
            self.report.warning(
                f"Provenance events could not be fetched for processor \
                    {processor.id} of type {processor.name}",
                self.config.site_url,
            )
            logger.warning(provenance_response.text)
        return

    def submit_provenance_query(self, processor, eventType, startDate, endDate):
        older_version: bool = self.nifi_flow.version is not None and version.parse(
            self.nifi_flow.version
        ) < version.parse("1.13.0")

        if older_version:
            searchTerms = {
                "ProcessorID": processor.id,
                "EventType": eventType,
            }
        else:
            searchTerms = {
                "ProcessorID": {"value": processor.id},  # type: ignore
                "EventType": {"value": eventType},  # type: ignore
            }

        payload = json.dumps(
            {
                "provenance": {
                    "request": {
                        "maxResults": 1000,
                        "summarize": False,
                        "searchTerms": searchTerms,
                        "startDate": startDate.strftime("%m/%d/%Y %H:%M:%S %Z"),
                        "endDate": (
                            endDate.strftime("%m/%d/%Y %H:%M:%S %Z")
                            if endDate
                            else None
                        ),
                    }
                }
            }
        )
        logger.debug(payload)
        self.session.headers.update({})

        self.session.headers.update({"Content-Type": "application/json"})
        provenance_response = self.session.post(
            url=urljoin(self.rest_api_base_url, PROVENANCE_ENDPOINT),
            data=payload,
        )

        # Revert to default content-type if basic-auth
        if self.config.auth is NifiAuthType.BASIC_AUTH:
            self.session.headers.update(
                {
                    "Content-Type": "application/x-www-form-urlencoded",
                }
            )

        return provenance_response

    def delete_provenance(self, provenance_uri):
        delete_response = self.session.delete(provenance_uri)
        if not delete_response.ok:
            logger.error("failed to delete provenance ", provenance_uri)

    def construct_workunits(self) -> Iterable[MetadataWorkUnit]:  # noqa: C901
        rootpg = self.nifi_flow.root_process_group
        flow_name = rootpg.name  # self.config.site_name
        flow_urn = self.make_flow_urn()
        flow_properties = {}
        if self.nifi_flow.clustered is not None:
            flow_properties["clustered"] = str(self.nifi_flow.clustered)
        if self.nifi_flow.version is not None:
            flow_properties["version"] = str(self.nifi_flow.version)
        yield self.construct_flow_workunits(
            flow_urn, flow_name, self.make_external_url(rootpg.id), flow_properties
        )

        for component in self.nifi_flow.components.values():
            job_name = component.name
            job_urn = builder.make_data_job_urn_with_flow(flow_urn, component.id)

            incoming = list(
                filter(lambda x: x[1] == component.id, self.nifi_flow.connections)
            )
            outgoing = list(
                filter(lambda x: x[0] == component.id, self.nifi_flow.connections)
            )
            inputJobs = set()
            jobProperties = None

            if component.nifi_type is NifiType.PROCESSOR:
                jobProperties = {
                    k: str(v)
                    for k, v in component.config.items()  # type: ignore
                    if k
                    in [
                        "schedulingPeriod",
                        "schedulingStrategy",
                        "executionNode",
                        "concurrentlySchedulableTaskCount",
                    ]
                }
                jobProperties["properties"] = json.dumps(
                    component.config.get("properties")  # type: ignore
                )
                if component.last_event_time is not None:
                    jobProperties["last_event_time"] = component.last_event_time

                for dataset in component.inlets.values():
                    yield from self.construct_dataset_workunits(
                        dataset.platform,
                        dataset.dataset_name,
                        dataset.dataset_urn,
                        datasetProperties=dataset.dataset_properties,
                    )

                for dataset in component.outlets.values():
                    yield from self.construct_dataset_workunits(
                        dataset.platform,
                        dataset.dataset_name,
                        dataset.dataset_urn,
                        datasetProperties=dataset.dataset_properties,
                    )

            for edge in incoming:
                incoming_from = edge[0]
                if incoming_from in self.nifi_flow.remotely_accessible_ports.keys():
                    dataset_name = f"{self.config.site_name}.{self.nifi_flow.remotely_accessible_ports[incoming_from].name}"
                    dataset_urn = builder.make_dataset_urn(
                        NIFI, dataset_name, self.config.env
                    )
                    component.inlets[dataset_urn] = ExternalDataset(
                        NIFI,
                        dataset_name,
                        dict(nifi_uri=self.config.site_url),
                        dataset_urn,
                    )
                else:
                    inputJobs.add(
                        builder.make_data_job_urn_with_flow(flow_urn, incoming_from)
                    )

            for edge in outgoing:
                outgoing_to = edge[1]
                if outgoing_to in self.nifi_flow.remotely_accessible_ports.keys():
                    dataset_name = f"{self.config.site_name}.{self.nifi_flow.remotely_accessible_ports[outgoing_to].name}"
                    dataset_urn = builder.make_dataset_urn(
                        NIFI, dataset_name, self.config.env
                    )
                    component.outlets[dataset_urn] = ExternalDataset(
                        NIFI,
                        dataset_name,
                        dict(nifi_uri=self.config.site_url),
                        dataset_urn,
                    )

            if component.nifi_type is NifiType.REMOTE_INPUT_PORT:
                # TODO - if target_uris is not set, but http proxy is used in RPG
                site_urls = component.target_uris.split(",")  # type: ignore
                for site_url in site_urls:
                    if site_url not in self.config.site_url_to_site_name:
                        self.report.warning(
                            f"Site with url {site_url} is being used in flow but\
                            corresponding site name is not configured via site_url_to_site_name.\
                            This may result in broken lineage.",
                            site_url,
                        )
                    else:
                        site_name = self.config.site_url_to_site_name[site_url]
                        dataset_name = f"{site_name}.{component.name}"
                        dataset_urn = builder.make_dataset_urn(
                            NIFI, dataset_name, self.config.env
                        )
                        component.outlets[dataset_urn] = ExternalDataset(
                            NIFI, dataset_name, dict(nifi_uri=site_url), dataset_urn
                        )
                        break

            if component.nifi_type is NifiType.REMOTE_OUTPUT_PORT:
                site_urls = component.target_uris.split(",")  # type: ignore
                for site_url in site_urls:
                    if site_url not in self.config.site_url_to_site_name:
                        self.report.warning(
                            f"Site with url {site_url} is being used in flow but\
                            corresponding site name is not configured via site_url_to_site_name.\
                            This may result in broken lineage.",
                            self.config.site_url,
                        )
                    else:
                        site_name = self.config.site_url_to_site_name[site_url]

                        dataset_name = f"{site_name}.{component.name}"
                        dataset_urn = builder.make_dataset_urn(
                            NIFI, dataset_name, self.config.env
                        )
                        component.inlets[dataset_urn] = ExternalDataset(
                            NIFI, dataset_name, dict(nifi_uri=site_url), dataset_urn
                        )
                        break

            if self.config.emit_process_group_as_container:
                # We emit process groups only for all nifi components qualifying as datajobs
                yield from self.construct_process_group_workunits(
                    component.parent_group_id
                )

            yield from self.construct_job_workunits(
                job_urn,
                job_name,
                component.parent_group_id,
                external_url=self.make_external_url(
                    component.parent_group_id, component.id, component.parent_rpg_id
                ),
                job_type=NIFI.upper() + "_" + component.nifi_type.value,
                description=component.comments,
                job_properties=jobProperties,
                inlets=list(component.inlets.keys()),
                outlets=list(component.outlets.keys()),
                inputJobs=list(inputJobs),
                status=component.status,
            )

        for port in self.nifi_flow.remotely_accessible_ports.values():
            dataset_name = f"{self.config.site_name}.{port.name}"
            dataset_platform = NIFI
            yield from self.construct_dataset_workunits(
                dataset_platform,
                dataset_name,
                external_url=self.make_external_url(port.parent_group_id, port.id),
            )

    def make_flow_urn(self) -> str:
        return builder.make_data_flow_urn(
            NIFI, self.nifi_flow.root_process_group.id, self.config.env
        )

    def process_provenance_events(self):
        startDate = datetime.now(timezone.utc) - timedelta(
            days=self.config.provenance_days
        )

        eventAnalyzer = NifiProcessorProvenanceEventAnalyzer()
        eventAnalyzer.env = self.config.env

        for component in self.nifi_flow.components.values():
            if component.nifi_type is NifiType.PROCESSOR:
                eventType = eventAnalyzer.KNOWN_INGRESS_EGRESS_PROCESORS[component.type]
                events = self.fetch_provenance_events(component, eventType, startDate)
                for event in events:
                    dataset = eventAnalyzer.provenance_event_to_lineage_map[
                        component.type
                    ](event)
                    if eventType in [
                        NifiEventType.CREATE,
                        NifiEventType.FETCH,
                        NifiEventType.RECEIVE,
                    ]:
                        component.inlets[dataset.dataset_urn] = dataset
                    else:
                        component.outlets[dataset.dataset_urn] = dataset

    def authenticate(self):
        if self.config.auth is NifiAuthType.NO_AUTH:
            # Token not required
            return

        if self.config.auth is NifiAuthType.BASIC_AUTH:
            assert self.config.username is not None
            assert self.config.password is not None
            self.session.auth = HTTPBasicAuth(
                self.config.username, self.config.password
            )
            self.session.headers.update(
                {
                    "Content-Type": "application/x-www-form-urlencoded",
                }
            )
            return

        if self.config.auth is NifiAuthType.CLIENT_CERT:
            self.session.mount(
                self.rest_api_base_url,
                SSLAdapter(
                    certfile=self.config.client_cert_file,
                    keyfile=self.config.client_key_file,
                    password=self.config.client_key_password,
                ),
            )
            return

        # get token flow
        token_response: Response
        if self.config.auth is NifiAuthType.SINGLE_USER:
            token_response = self.session.post(
                url=urljoin(self.rest_api_base_url, TOKEN_ENDPOINT),
                data={
                    "username": self.config.username,
                    "password": self.config.password,
                },
            )
        elif self.config.auth is NifiAuthType.KERBEROS:
            token_response = self.session.post(
                url=urljoin(self.rest_api_base_url, KERBEROS_TOKEN_ENDPOINT),
                auth=HTTPSPNEGOAuth(),
            )
        else:
            raise Exception(f"Unsupported auth: {self.config.auth}")

        token_response.raise_for_status()
        self.session.headers.update({"Authorization": "Bearer " + token_response.text})

    def get_workunits_internal(self) -> Iterable[MetadataWorkUnit]:
        try:
            self.authenticate()
        except Exception as e:
            self.report.failure("Failed to authenticate", self.config.site_url, exc=e)
            return

        # Creates nifi_flow by invoking /flow rest api and saves as self.nifi_flow
        try:
            self.create_nifi_flow()
        except Exception as e:
            self.report.failure(
                "Failed to get root process group flow", self.config.site_url, exc=e
            )
            return

        # Updates inlets and outlets of nifi_flow.components by invoking /provenance rest api
        self.process_provenance_events()

        # Reads and translates entities from self.nifi_flow into mcps
        yield from self.construct_workunits()

    def make_external_url(
        self,
        parent_group_id: str,
        component_id: Optional[str] = "",
        parent_rpg_id: Optional[str] = None,
    ) -> str:
        if parent_rpg_id is not None:
            component_id = parent_rpg_id
        return urljoin(
            self.config.site_url,
            f"?processGroupId={parent_group_id}&componentIds={component_id}",
        )

    def construct_flow_workunits(
        self,
        flow_urn: str,
        flow_name: str,
        external_url: str,
        flow_properties: Optional[Dict[str, str]] = None,
    ) -> MetadataWorkUnit:
        return MetadataChangeProposalWrapper(
            entityUrn=flow_urn,
            aspect=DataFlowInfoClass(
                name=flow_name,
                customProperties=flow_properties,
                externalUrl=external_url,
            ),
        ).as_workunit()

    def construct_job_workunits(
        self,
        job_urn: str,
        job_name: str,
        parent_group_id: str,
        external_url: str,
        job_type: str,
        description: Optional[str],
        job_properties: Optional[Dict[str, str]] = None,
        inlets: List[str] = [],
        outlets: List[str] = [],
        inputJobs: List[str] = [],
        status: Optional[str] = None,
    ) -> Iterable[MetadataWorkUnit]:
        if job_properties:
            job_properties = {k: v for k, v in job_properties.items() if v is not None}

        yield MetadataChangeProposalWrapper(
            entityUrn=job_urn,
            aspect=DataJobInfoClass(
                name=job_name,
                type=job_type,
                description=description,
                customProperties=job_properties,
                externalUrl=external_url,
                status=status,
            ),
        ).as_workunit()

        # If dataJob had container aspect, we would ideally only emit it
        # and browse path v2 would automatically be generated.
        yield self.gen_browse_path_v2_workunit(job_urn, parent_group_id)

        inlets.sort()
        outlets.sort()
        inputJobs.sort()

        if self.config.incremental_lineage:
            patch_builder: DataJobPatchBuilder = DataJobPatchBuilder(job_urn)
            for inlet in inlets:
                patch_builder.add_input_dataset(inlet)
            for outlet in outlets:
                patch_builder.add_output_dataset(outlet)
            for inJob in inputJobs:
                patch_builder.add_input_datajob(inJob)
            for patch_mcp in patch_builder.build():
                yield MetadataWorkUnit(
                    id=f"{job_urn}-{patch_mcp.aspectName}", mcp_raw=patch_mcp
                )
        else:
            yield MetadataChangeProposalWrapper(
                entityUrn=job_urn,
                aspect=DataJobInputOutputClass(
                    inputDatasets=inlets,
                    outputDatasets=outlets,
                    inputDatajobs=inputJobs,
                ),
            ).as_workunit()

    def gen_browse_path_v2_workunit(
        self, entity_urn: str, process_group_id: str
    ) -> MetadataWorkUnit:
        flow_urn = self.make_flow_urn()
        return MetadataChangeProposalWrapper(
            entityUrn=entity_urn,
            aspect=BrowsePathsV2Class(
                path=[
                    BrowsePathEntryClass(id=flow_urn, urn=flow_urn),
                    *self._get_browse_path_v2_entries(process_group_id),
                ]
            ),
        ).as_workunit()

    def _get_browse_path_v2_entries(
        self, process_group_id: str
    ) -> List[BrowsePathEntryClass]:
        """Browse path entries till current process group"""
        if self._is_root_process_group(process_group_id):
            return []

        current_process_group = self.nifi_flow.processGroups[process_group_id]
        assert (
            current_process_group.parent_group_id
        )  # always present for non-root process group
        parent_browse_path = self._get_browse_path_v2_entries(
            current_process_group.parent_group_id
        )

        if self.config.emit_process_group_as_container:
            container_urn = self.gen_process_group_key(process_group_id).as_urn()
            current_browse_entry = BrowsePathEntryClass(
                id=container_urn, urn=container_urn
            )
        else:
            current_browse_entry = BrowsePathEntryClass(id=current_process_group.name)
        return parent_browse_path + [current_browse_entry]

    def _is_root_process_group(self, process_group_id: str) -> bool:
        return self.nifi_flow.root_process_group.id == process_group_id

    def construct_process_group_workunits(
        self, process_group_id: str
    ) -> Iterable[MetadataWorkUnit]:
        if (
            self._is_root_process_group(process_group_id)
            or process_group_id in self.processed_pgs
        ):
            return
        self.processed_pgs.append(process_group_id)

        pg = self.nifi_flow.processGroups[process_group_id]
        container_key = self.gen_process_group_key(process_group_id)
        yield from gen_containers(
            container_key=container_key,
            name=pg.name,
            sub_types=[JobContainerSubTypes.NIFI_PROCESS_GROUP],
            parent_container_key=(
                self.gen_process_group_key(pg.parent_group_id)
                if pg.parent_group_id
                and not self._is_root_process_group(pg.parent_group_id)
                else None
            ),
        )

        if pg.parent_group_id:  # always true for non-root process group
            yield from self.construct_process_group_workunits(pg.parent_group_id)

            if self._is_root_process_group(pg.parent_group_id):
                yield self.gen_browse_path_v2_workunit(
                    container_key.as_urn(), pg.parent_group_id
                )

    def gen_process_group_key(self, process_group_id: str) -> ProcessGroupKey:
        return ProcessGroupKey(
            process_group_id=process_group_id, platform=NIFI, env=self.config.env
        )

    def construct_dataset_workunits(
        self,
        dataset_platform: str,
        dataset_name: str,
        dataset_urn: Optional[str] = None,
        external_url: Optional[str] = None,
        datasetProperties: Optional[Dict[str, str]] = None,
    ) -> Iterable[MetadataWorkUnit]:
        if not dataset_urn:
            dataset_urn = builder.make_dataset_urn(
                dataset_platform, dataset_name, self.config.env
            )

        yield MetadataChangeProposalWrapper(
            entityUrn=dataset_urn,
            aspect=DataPlatformInstanceClass(
                platform=builder.make_data_platform_urn(dataset_platform)
            ),
        ).as_workunit()

        yield MetadataChangeProposalWrapper(
            entityUrn=dataset_urn,
            aspect=DatasetPropertiesClass(
                externalUrl=external_url, customProperties=datasetProperties
            ),
        ).as_workunit()



from typing import Iterable, List, Set, TypeVar

_T = TypeVar("_T")


def deduplicate_list(iterable: Iterable[_T]) -> List[_T]:
    """
    Remove duplicates from an iterable, preserving order.
    This serves as a replacement for OrderedSet, which is broken in Python 3.10.
    """

    seen: Set[_T] = set()
    result: List[_T] = []
    for item in iterable:
        if item not in seen:
            seen.add(item)
            result.append(item)
    return result



import json
import logging
from datetime import datetime
from typing import Any, Generic, Iterable, List, Optional, Tuple, TypeVar

from sqlalchemy import create_engine
from sqlalchemy.engine import Row
from typing_extensions import Protocol

from datahub.emitter.aspect import ASPECT_MAP
from datahub.emitter.mcp import MetadataChangeProposalWrapper
from datahub.emitter.serialization_helper import post_json_transform
from datahub.ingestion.source.datahub.config import DataHubSourceConfig
from datahub.ingestion.source.datahub.report import DataHubSourceReport
from datahub.ingestion.source.sql.sql_config import SQLAlchemyConnectionConfig
from datahub.metadata.schema_classes import ChangeTypeClass, SystemMetadataClass
from datahub.utilities.lossy_collections import LossyDict, LossyList

logger = logging.getLogger(__name__)

# Should work for at least mysql, mariadb, postgres
DATETIME_FORMAT = "%Y-%m-%d %H:%M:%S.%f"


class VersionOrderable(Protocol):
    createdon: Any  # Should restrict to only orderable types
    version: int


ROW = TypeVar("ROW", bound=VersionOrderable)


class VersionOrderer(Generic[ROW]):
    """Orders rows by (createdon, version == 0).

    That is, orders rows first by createdon, and for equal timestamps, puts version 0 rows last.
    """

    def __init__(self, enabled: bool):
        # Stores all version 0 aspects for a given createdon timestamp
        # Once we have emitted all aspects for a given timestamp, we can emit the version 0 aspects
        # Guaranteeing that, for a given timestamp, we always ingest version 0 aspects last
        self.queue: Optional[Tuple[datetime, List[ROW]]] = None
        self.enabled = enabled

    def __call__(self, rows: Iterable[ROW]) -> Iterable[ROW]:
        for row in rows:
            yield from self._process_row(row)
        yield from self._flush_queue()

    def _process_row(self, row: ROW) -> Iterable[ROW]:
        if not self.enabled:
            yield row
            return

        yield from self._attempt_queue_flush(row)
        if row.version == 0:
            self._add_to_queue(row)
        else:
            yield row

    def _add_to_queue(self, row: ROW) -> None:
        if self.queue is None:
            self.queue = (row.createdon, [row])
        else:
            self.queue[1].append(row)

    def _attempt_queue_flush(self, row: ROW) -> Iterable[ROW]:
        if self.queue is None:
            return

        if row.createdon > self.queue[0]:
            yield from self._flush_queue()

    def _flush_queue(self) -> Iterable[ROW]:
        if self.queue is not None:
            yield from self.queue[1]
            self.queue = None


class DataHubDatabaseReader:
    def __init__(
        self,
        config: DataHubSourceConfig,
        connection_config: SQLAlchemyConnectionConfig,
        report: DataHubSourceReport,
    ):
        self.config = config
        self.report = report
        self.engine = create_engine(
            url=connection_config.get_sql_alchemy_url(),
            **connection_config.options,
        )

    @property
    def query(self) -> str:
        # May repeat rows for the same date
        # Offset is generally 0, unless we repeat the same createdon twice

        # Ensures stable order, chronological per (urn, aspect)
        # Relies on createdon order to reflect version order
        # Ordering of entries with the same createdon is handled by VersionOrderer
        return f"""
            SELECT urn, aspect, metadata, systemmetadata, createdon, version
            FROM {self.engine.dialect.identifier_preparer.quote(self.config.database_table_name)}
            WHERE createdon >= %(since_createdon)s
            {"" if self.config.include_all_versions else "AND version = 0"}
            ORDER BY createdon, urn, aspect, version
            LIMIT %(limit)s
            OFFSET %(offset)s
        """

    def get_aspects(
        self, from_createdon: datetime, stop_time: datetime
    ) -> Iterable[Tuple[MetadataChangeProposalWrapper, datetime]]:
        orderer = VersionOrderer[Row](enabled=self.config.include_all_versions)
        rows = self._get_rows(from_createdon=from_createdon, stop_time=stop_time)
        for row in orderer(rows):
            mcp = self._parse_row(row)
            if mcp:
                yield mcp, row.createdon

    def _get_rows(self, from_createdon: datetime, stop_time: datetime) -> Iterable[Row]:
        with self.engine.connect() as conn:
            ts = from_createdon
            offset = 0
            while ts.timestamp() <= stop_time.timestamp():
                logger.debug(f"Polling database aspects from {ts}")
                rows = conn.execute(
                    self.query,
                    since_createdon=ts.strftime(DATETIME_FORMAT),
                    limit=self.config.database_query_batch_size,
                    offset=offset,
                )
                if not rows.rowcount:
                    return

                for i, row in enumerate(rows):
                    yield row

                if ts == row.createdon:
                    offset += i + 1
                else:
                    ts = row.createdon
                    offset = 0

    def _parse_row(self, row: Row) -> Optional[MetadataChangeProposalWrapper]:
        try:
            json_aspect = post_json_transform(json.loads(row.metadata))
            json_metadata = post_json_transform(json.loads(row.systemmetadata or "{}"))
            system_metadata = SystemMetadataClass.from_obj(json_metadata)
            return MetadataChangeProposalWrapper(
                entityUrn=row.urn,
                aspect=ASPECT_MAP[row.aspect].from_obj(json_aspect),
                systemMetadata=system_metadata,
                changeType=ChangeTypeClass.UPSERT,
            )
        except Exception as e:
            logger.warning(
                f"Failed to parse metadata for {row.urn}: {e}", exc_info=True
            )
            self.report.num_database_parse_errors += 1
            self.report.database_parse_errors.setdefault(
                str(e), LossyDict()
            ).setdefault(row.aspect, LossyList()).append(row.urn)
            return None



import logging
from enum import Enum
from pathlib import Path
from typing import List, Optional

import yaml
from pydantic import validator
from ruamel.yaml import YAML

from datahub.configuration.common import ConfigModel
from datahub.emitter.mcp import MetadataChangeProposalWrapper
from datahub.ingestion.graph.client import DataHubGraph, get_default_graph
from datahub.metadata.schema_classes import (
    PropertyValueClass,
    StructuredPropertyDefinitionClass,
)
from datahub.utilities.urns.urn import Urn

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


class AllowedTypes(Enum):
    STRING = "string"
    RICH_TEXT = "rich_text"
    NUMBER = "number"
    DATE = "date"
    URN = "urn"

    @staticmethod
    def check_allowed_type(value: str) -> bool:
        return value in [allowed_type.value for allowed_type in AllowedTypes]

    @staticmethod
    def values():
        return ", ".join([allowed_type.value for allowed_type in AllowedTypes])


class AllowedValue(ConfigModel):
    value: str
    description: Optional[str] = None


class TypeQualifierAllowedTypes(ConfigModel):
    allowed_types: List[str]

    @validator("allowed_types")
    def validate_allowed_types(cls, v):
        validated_entity_type_urns = []
        if v:
            with get_default_graph() as graph:
                for et in v:
                    validated_urn = Urn.make_entity_type_urn(et)
                    if graph.exists(validated_urn):
                        validated_entity_type_urns.append(validated_urn)
                    else:
                        logger.warn(
                            f"Input {et} is not a valid entity type urn. Skipping."
                        )
        v = validated_entity_type_urns
        if not v:
            logger.warn("No allowed_types given within type_qualifier.")
        return v


class StructuredProperties(ConfigModel):
    id: Optional[str] = None
    urn: Optional[str] = None
    qualified_name: Optional[str] = None
    type: str
    value_entity_types: Optional[List[str]] = None
    description: Optional[str] = None
    display_name: Optional[str] = None
    entity_types: Optional[List[str]] = None
    cardinality: Optional[str] = None
    allowed_values: Optional[List[AllowedValue]] = None
    type_qualifier: Optional[TypeQualifierAllowedTypes] = None
    immutable: Optional[bool] = False

    @property
    def fqn(self) -> str:
        assert self.urn is not None
        return (
            self.qualified_name
            or self.id
            or Urn.create_from_string(self.urn).get_entity_id()[0]
        )

    @validator("urn", pre=True, always=True)
    def urn_must_be_present(cls, v, values):
        if not v:
            assert "id" in values, "id must be present if urn is not"
            return f"urn:li:structuredProperty:{values['id']}"
        return v

    @staticmethod
    def create(file: str) -> None:
        emitter: DataHubGraph

        with get_default_graph() as emitter:
            with open(file) as fp:
                structuredproperties: List[dict] = yaml.safe_load(fp)
                for structuredproperty_raw in structuredproperties:
                    structuredproperty = StructuredProperties.parse_obj(
                        structuredproperty_raw
                    )
                    if not structuredproperty.type.islower():
                        structuredproperty.type = structuredproperty.type.lower()
                        logger.warn(
                            f"Structured property type should be lowercase. Updated to {structuredproperty.type}"
                        )
                    if not AllowedTypes.check_allowed_type(structuredproperty.type):
                        raise ValueError(
                            f"Type {structuredproperty.type} is not allowed. Allowed types are {AllowedTypes.values()}"
                        )
                    mcp = MetadataChangeProposalWrapper(
                        entityUrn=structuredproperty.urn,
                        aspect=StructuredPropertyDefinitionClass(
                            qualifiedName=structuredproperty.fqn,
                            valueType=Urn.make_data_type_urn(structuredproperty.type),
                            displayName=structuredproperty.display_name,
                            description=structuredproperty.description,
                            entityTypes=[
                                Urn.make_entity_type_urn(entity_type)
                                for entity_type in structuredproperty.entity_types or []
                            ],
                            cardinality=structuredproperty.cardinality,
                            immutable=structuredproperty.immutable,
                            allowedValues=(
                                [
                                    PropertyValueClass(
                                        value=v.value, description=v.description
                                    )
                                    for v in structuredproperty.allowed_values
                                ]
                                if structuredproperty.allowed_values
                                else None
                            ),
                            typeQualifier=(
                                {
                                    "allowedTypes": structuredproperty.type_qualifier.allowed_types
                                }
                                if structuredproperty.type_qualifier
                                else None
                            ),
                        ),
                    )
                    emitter.emit_mcp(mcp)

                    logger.info(f"Created structured property {structuredproperty.urn}")

    @classmethod
    def from_datahub(cls, graph: DataHubGraph, urn: str) -> "StructuredProperties":

        structured_property: Optional[
            StructuredPropertyDefinitionClass
        ] = graph.get_aspect(urn, StructuredPropertyDefinitionClass)
        assert structured_property is not None
        return StructuredProperties(
            urn=urn,
            qualified_name=structured_property.qualifiedName,
            display_name=structured_property.displayName,
            type=structured_property.valueType,
            description=structured_property.description,
            entity_types=structured_property.entityTypes,
            cardinality=structured_property.cardinality,
            allowed_values=(
                [
                    AllowedValue(
                        value=av.value,
                        description=av.description,
                    )
                    for av in structured_property.allowedValues or []
                ]
                if structured_property.allowedValues is not None
                else None
            ),
            type_qualifier=(
                {"allowed_types": structured_property.typeQualifier.get("allowedTypes")}
                if structured_property.typeQualifier
                else None
            ),
        )

    def to_yaml(
        self,
        file: Path,
    ) -> None:
        with open(file, "w") as fp:
            yaml = YAML(typ="rt")  # default, if not specfied, is 'rt' (round-trip)
            yaml.indent(mapping=2, sequence=4, offset=2)
            yaml.default_flow_style = False
            yaml.dump(self.dict(), fp)



from collections.abc import Iterator
from dataclasses import dataclass
from typing import Any, Iterable
from urllib.parse import urlparse

import boto3
import smart_open

from datahub.ingestion.fs import s3_fs
from datahub.ingestion.fs.fs_base import FileInfo, FileSystem


def parse_s3_path(path: str) -> "S3Path":
    parsed = urlparse(path)
    return S3Path(parsed.netloc, parsed.path.lstrip("/"))


def assert_ok_status(s3_response):
    is_ok = s3_response["ResponseMetadata"]["HTTPStatusCode"] == 200
    assert (
        is_ok
    ), f"Failed to fetch S3 object, error message: {s3_response['Error']['Message']}"


@dataclass
class S3Path:
    bucket: str
    key: str

    def __str__(self):
        return f"S3Path({self.bucket}, {self.key})"


class S3ListIterator(Iterator):

    MAX_KEYS = 1000

    def __init__(
        self, s3_client: Any, bucket: str, prefix: str, max_keys: int = MAX_KEYS
    ) -> None:
        self._s3 = s3_client
        self._bucket = bucket
        self._prefix = prefix
        self._max_keys = max_keys
        self._file_statuses: Iterator = iter([])
        self._token = ""
        self.fetch()

    def __next__(self) -> FileInfo:
        try:
            return next(self._file_statuses)
        except StopIteration:
            if self._token:
                self.fetch()
                return next(self._file_statuses)
            else:
                raise StopIteration()

    def fetch(self):
        params = dict(Bucket=self._bucket, Prefix=self._prefix, MaxKeys=self._max_keys)
        if self._token:
            params.update(ContinuationToken=self._token)

        response = self._s3.list_objects_v2(**params)

        s3_fs.assert_ok_status(response)

        self._file_statuses = iter(
            [
                FileInfo(f"s3://{response['Name']}/{x['Key']}", x["Size"], is_file=True)
                for x in response.get("Contents", [])
            ]
        )
        self._token = response.get("NextContinuationToken")


class S3FileSystem(FileSystem):
    def __init__(self, **kwargs):
        self.s3 = boto3.client("s3", **kwargs)

    @classmethod
    def create(cls, **kwargs):
        return S3FileSystem(**kwargs)

    def open(self, path: str, **kwargs: Any) -> Any:
        transport_params = kwargs.update({"client": self.s3})
        return smart_open.open(path, mode="rb", transport_params=transport_params)

    def file_status(self, path: str) -> FileInfo:
        s3_path = parse_s3_path(path)
        try:
            response = self.s3.get_object_attributes(
                Bucket=s3_path.bucket, Key=s3_path.key, ObjectAttributes=["ObjectSize"]
            )
            assert_ok_status(response)
            return FileInfo(path, response["ObjectSize"], is_file=True)
        except Exception as e:
            if (
                hasattr(e, "response")
                and e.response["ResponseMetadata"]["HTTPStatusCode"] == 404
            ):
                return FileInfo(path, 0, is_file=False)
            else:
                raise e

    def list(self, path: str) -> Iterable[FileInfo]:
        s3_path = parse_s3_path(path)
        return S3ListIterator(self.s3, s3_path.bucket, s3_path.key)



import logging
import re
from abc import ABC, abstractmethod
from functools import lru_cache
from typing import Dict, List, Optional

from datahub.emitter.mce_builder import make_dataset_urn_with_platform_instance
from datahub.ingestion.api.common import PipelineContext
from datahub.ingestion.source.looker.looker_common import (
    LookerExplore,
    LookerViewId,
    ViewField,
    ViewFieldType,
)
from datahub.ingestion.source.looker.looker_view_id_cache import LookerViewIdCache
from datahub.ingestion.source.looker.lookml_concept_context import (
    LookerFieldContext,
    LookerViewContext,
)
from datahub.ingestion.source.looker.lookml_config import (
    DERIVED_VIEW_SUFFIX,
    NAME,
    LookerConnectionDefinition,
    LookMLSourceConfig,
    LookMLSourceReport,
)
from datahub.ingestion.source.looker.urn_functions import get_qualified_table_name
from datahub.sql_parsing.sqlglot_lineage import (
    ColumnLineageInfo,
    ColumnRef,
    SqlParsingResult,
    Urn,
    create_lineage_sql_parsed_result,
)

logger = logging.getLogger(__name__)


def is_derived_view(view_name: str) -> bool:
    if DERIVED_VIEW_SUFFIX in view_name.lower():
        return True

    return False


def get_derived_looker_view_id(
    qualified_table_name: str,
    looker_view_id_cache: "LookerViewIdCache",
    base_folder_path: str,
) -> Optional[LookerViewId]:
    # qualified_table_name can be in either of below format
    # 1) db.schema.employee_income_source.sql_table_name
    # 2) db.employee_income_source.sql_table_name
    # 3) employee_income_source.sql_table_name
    # In any of the form we need the text coming before ".sql_table_name" and after last "."
    parts: List[str] = re.split(
        DERIVED_VIEW_SUFFIX, qualified_table_name, flags=re.IGNORECASE
    )
    view_name: str = parts[0].split(".")[-1]

    looker_view_id: Optional[LookerViewId] = looker_view_id_cache.get_looker_view_id(
        view_name=view_name,
        base_folder_path=base_folder_path,
    )

    return looker_view_id


def resolve_derived_view_urn_of_col_ref(
    column_refs: List[ColumnRef],
    looker_view_id_cache: "LookerViewIdCache",
    base_folder_path: str,
    config: LookMLSourceConfig,
) -> List[ColumnRef]:

    new_column_refs: List[ColumnRef] = []
    for col_ref in column_refs:
        if is_derived_view(col_ref.table.lower()):
            new_urns: List[str] = fix_derived_view_urn(
                urns=[col_ref.table],
                looker_view_id_cache=looker_view_id_cache,
                base_folder_path=base_folder_path,
                config=config,
            )
            if not new_urns:
                logger.warning(
                    f"Not able to resolve to derived view looker id for {col_ref.table}"
                )
                continue

            new_column_refs.append(ColumnRef(table=new_urns[0], column=col_ref.column))
        else:
            new_column_refs.append(col_ref)

    return new_column_refs


def fix_derived_view_urn(
    urns: List[str],
    looker_view_id_cache: "LookerViewIdCache",
    base_folder_path: str,
    config: LookMLSourceConfig,
) -> List[str]:
    # Regenerate view urn if .sql_table_name is present in urn
    new_urns: List[str] = []
    for urn in urns:
        if is_derived_view(urn):
            looker_view_id = get_derived_looker_view_id(
                qualified_table_name=get_qualified_table_name(urn),
                looker_view_id_cache=looker_view_id_cache,
                base_folder_path=base_folder_path,
            )

            if looker_view_id is None:
                logger.warning(
                    f"Not able to resolve to derived view looker id for {urn}"
                )
                continue

            new_urns.append(looker_view_id.get_urn(config=config))
        else:
            new_urns.append(urn)

    return new_urns


def _platform_names_have_2_parts(platform: str) -> bool:
    return platform in {"hive", "mysql", "athena"}


def _drop_hive_dot(urn: str) -> str:
    """
    This is special handling for hive platform where "hive." is coming in urn's id because of the way SQL
    is written in lookml.

    Example: urn:li:dataset:(urn:li:dataPlatform:hive,hive.my_database.my_table,PROD)

    Here we need to transform hive.my_database.my_table to my_database.my_table
    """
    if urn.startswith("urn:li:dataset:(urn:li:dataPlatform:hive"):
        return re.sub(r"hive\.", "", urn)

    return urn


def _drop_hive_dot_from_upstream(upstreams: List[ColumnRef]) -> List[ColumnRef]:
    return [
        ColumnRef(table=_drop_hive_dot(column_ref.table), column=column_ref.column)
        for column_ref in upstreams
    ]


def _generate_fully_qualified_name(
    sql_table_name: str,
    connection_def: LookerConnectionDefinition,
    reporter: LookMLSourceReport,
) -> str:
    """Returns a fully qualified dataset name, resolved through a connection definition.
    Input sql_table_name can be in three forms: table, db.table, db.schema.table"""
    # TODO: This function should be extracted out into a Platform specific naming class since name translations
    #  are required across all connectors

    # Bigquery has "project.db.table" which can be mapped to db.schema.table form
    # All other relational db's follow "db.schema.table"
    # With the exception of mysql, hive, athena which are "db.table"

    # first detect which one we have
    parts = len(sql_table_name.split("."))

    if parts == 3:
        # fully qualified, but if platform is of 2-part, we drop the first level
        if _platform_names_have_2_parts(connection_def.platform):
            sql_table_name = ".".join(sql_table_name.split(".")[1:])
        return sql_table_name.lower()

    if parts == 1:
        # Bare table form
        if _platform_names_have_2_parts(connection_def.platform):
            dataset_name = f"{connection_def.default_db}.{sql_table_name}"
        else:
            dataset_name = f"{connection_def.default_db}.{connection_def.default_schema}.{sql_table_name}"
        return dataset_name.lower()

    if parts == 2:
        # if this is a 2 part platform, we are fine
        if _platform_names_have_2_parts(connection_def.platform):
            return sql_table_name.lower()
        # otherwise we attach the default top-level container
        dataset_name = f"{connection_def.default_db}.{sql_table_name}"
        return dataset_name.lower()

    reporter.report_warning(
        title="Malformed Table Name",
        message="Table name has more than 3 parts.",
        context=f"Table Name: {sql_table_name}",
    )
    return sql_table_name.lower()


class AbstractViewUpstream(ABC):
    """
    Implementation of this interface extracts the view upstream as per the way the view is bound to datasets.
    For detail explanation please refer lookml_concept_context.LookerViewContext documentation.
    """

    view_context: LookerViewContext
    looker_view_id_cache: LookerViewIdCache
    config: LookMLSourceConfig
    reporter: LookMLSourceReport
    ctx: PipelineContext

    def __init__(
        self,
        view_context: LookerViewContext,
        looker_view_id_cache: LookerViewIdCache,
        config: LookMLSourceConfig,
        reporter: LookMLSourceReport,
        ctx: PipelineContext,
    ):
        self.view_context = view_context
        self.looker_view_id_cache = looker_view_id_cache
        self.config = config
        self.reporter = reporter
        self.ctx = ctx

    @abstractmethod
    def get_upstream_column_ref(
        self, field_context: LookerFieldContext
    ) -> List[ColumnRef]:
        pass

    @abstractmethod
    def get_upstream_dataset_urn(self) -> List[Urn]:
        pass

    def create_fields(self) -> List[ViewField]:
        return []  # it is for the special case


class SqlBasedDerivedViewUpstream(AbstractViewUpstream, ABC):
    """
    Handle the case where upstream dataset is defined in derived_table.sql
    """

    def __init__(
        self,
        view_context: LookerViewContext,
        looker_view_id_cache: LookerViewIdCache,
        config: LookMLSourceConfig,
        reporter: LookMLSourceReport,
        ctx: PipelineContext,
    ):
        super().__init__(view_context, looker_view_id_cache, config, reporter, ctx)
        # These are the function where we need to catch the response once calculated
        self._get_spr = lru_cache(maxsize=1)(self.__get_spr)
        self._get_upstream_dataset_urn = lru_cache(maxsize=1)(
            self.__get_upstream_dataset_urn
        )

    def __get_spr(self) -> Optional[SqlParsingResult]:
        # for backward compatibility
        if not self.config.parse_table_names_from_sql:
            return None

        spr = create_lineage_sql_parsed_result(
            query=self.get_sql_query(),
            default_schema=self.view_context.view_connection.default_schema,
            default_db=self.view_context.view_connection.default_db,
            platform=self.view_context.view_connection.platform,
            platform_instance=self.view_context.view_connection.platform_instance,
            env=self.view_context.view_connection.platform_env or self.config.env,
            graph=self.ctx.graph,
        )
        return spr

    def __get_upstream_dataset_urn(self) -> List[Urn]:
        sql_parsing_result: Optional[SqlParsingResult] = self._get_spr()

        if sql_parsing_result is None:
            return []

        if sql_parsing_result.debug_info.table_error is not None:
            self.reporter.report_warning(
                title="Table Level Lineage Missing",
                message="Error in parsing derived sql",
                context=f"View-name: {self.view_context.name()}",
                exc=sql_parsing_result.debug_info.table_error,
            )
            return []

        upstream_dataset_urns: List[str] = [
            _drop_hive_dot(urn) for urn in sql_parsing_result.in_tables
        ]

        # fix any derived view reference present in urn
        upstream_dataset_urns = fix_derived_view_urn(
            urns=upstream_dataset_urns,
            looker_view_id_cache=self.looker_view_id_cache,
            base_folder_path=self.view_context.base_folder_path,
            config=self.config,
        )

        return upstream_dataset_urns

    def create_fields(self) -> List[ViewField]:
        spr: Optional[SqlParsingResult] = self._get_spr()

        if spr is None:
            return []

        if spr.debug_info.column_error is not None:
            self.reporter.report_warning(
                title="Column Level Lineage Missing",
                message="Error in parsing derived sql for CLL",
                context=f"View-name: {self.view_context.name()}",
                exc=spr.debug_info.column_error,
            )
            return []

        fields: List[ViewField] = []

        column_lineages: List[ColumnLineageInfo] = (
            spr.column_lineage if spr.column_lineage is not None else []
        )

        for cll in column_lineages:
            fields.append(
                ViewField(
                    name=cll.downstream.column,
                    label="",
                    type=(
                        cll.downstream.native_column_type
                        if cll.downstream.native_column_type is not None
                        else "unknown"
                    ),
                    description="",
                    field_type=ViewFieldType.UNKNOWN,
                    upstream_fields=_drop_hive_dot_from_upstream(cll.upstreams),
                )
            )

        return fields

    def get_upstream_column_ref(
        self, field_context: LookerFieldContext
    ) -> List[ColumnRef]:
        sql_parsing_result: Optional[SqlParsingResult] = self._get_spr()

        if sql_parsing_result is None:
            return []

        if sql_parsing_result.debug_info.column_error is not None:
            self.reporter.report_warning(
                title="Column Level Lineage Missing",
                message="Error in parsing derived sql for CLL",
                context=f"View-name: {self.view_context.name()}. "
                f"Error: {sql_parsing_result.debug_info.column_error}",
            )
            return []

        upstreams_column_refs: List[ColumnRef] = []
        if sql_parsing_result.column_lineage:
            for cll in sql_parsing_result.column_lineage:
                if cll.downstream.column == field_context.name():
                    upstreams_column_refs = cll.upstreams
                    break

        # field might get skip either because of Parser not able to identify the column from GMS
        # in-case of "select * from look_ml_view.SQL_TABLE_NAME" or extra field are defined in the looker view which is
        # referring to upstream table
        if self._get_upstream_dataset_urn() and not upstreams_column_refs:
            upstreams_column_refs = [
                ColumnRef(
                    table=self._get_upstream_dataset_urn()[
                        0
                    ],  # 0th index has table of from clause
                    column=column,
                )
                for column in field_context.column_name_in_sql_attribute()
            ]

        # fix any derived view reference present in urn
        upstreams_column_refs = resolve_derived_view_urn_of_col_ref(
            column_refs=upstreams_column_refs,
            looker_view_id_cache=self.looker_view_id_cache,
            base_folder_path=self.view_context.base_folder_path,
            config=self.config,
        )

        return _drop_hive_dot_from_upstream(upstreams_column_refs)

    def get_upstream_dataset_urn(self) -> List[Urn]:
        return self._get_upstream_dataset_urn()

    @abstractmethod
    def get_sql_query(self) -> str:
        pass


class DirectQueryUpstreamSource(SqlBasedDerivedViewUpstream):
    """
    Pattern 7 as per view-context documentation
    """

    def get_sql_query(self) -> str:
        return self.view_context.datahub_transformed_sql_table_name()


class DerivedQueryUpstreamSource(SqlBasedDerivedViewUpstream):
    """
    Pattern 4 as per view-context documentation
    """

    def get_sql_query(self) -> str:
        return self.view_context.datahub_transformed_sql()


class NativeDerivedViewUpstream(AbstractViewUpstream):
    """
    Handle the case where upstream dataset is defined as derived_table.explore_source
    """

    upstream_dataset_urns: List[str]
    explore_column_mapping: Dict

    def __init__(
        self,
        view_context: LookerViewContext,
        looker_view_id_cache: LookerViewIdCache,
        config: LookMLSourceConfig,
        reporter: LookMLSourceReport,
        ctx: PipelineContext,
    ):
        super().__init__(view_context, looker_view_id_cache, config, reporter, ctx)

        self._get_upstream_dataset_urn = lru_cache(maxsize=1)(
            self.__get_upstream_dataset_urn
        )
        self._get_explore_column_mapping = lru_cache(maxsize=1)(
            self.__get_explore_column_mapping
        )

    def __get_upstream_dataset_urn(self) -> List[str]:
        current_view_id: Optional[
            LookerViewId
        ] = self.looker_view_id_cache.get_looker_view_id(
            view_name=self.view_context.name(),
            base_folder_path=self.view_context.base_folder_path,
        )

        # Current view will always be present in cache. assert  will silence the lint
        assert current_view_id

        # We're creating a "LookerExplore" just to use the urn generator.
        upstream_dataset_urns: List[str] = [
            LookerExplore(
                name=self.view_context.explore_source()[NAME],
                model_name=current_view_id.model_name,
            ).get_explore_urn(self.config)
        ]

        return upstream_dataset_urns

    def __get_explore_column_mapping(self) -> Dict:
        explore_columns: Dict = self.view_context.explore_source().get("columns", {})

        explore_column_mapping = {}

        for column in explore_columns:
            explore_column_mapping[column[NAME]] = column

        return explore_column_mapping

    def get_upstream_column_ref(
        self, field_context: LookerFieldContext
    ) -> List[ColumnRef]:
        upstream_column_refs: List[ColumnRef] = []

        if not self._get_upstream_dataset_urn():
            # No upstream explore dataset found
            logging.debug(
                f"upstream explore not found for field {field_context.name()} of view {self.view_context.name()}"
            )
            return upstream_column_refs

        explore_urn: str = self._get_upstream_dataset_urn()[0]

        for column in field_context.column_name_in_sql_attribute():
            if column in self._get_explore_column_mapping():
                explore_column: Dict = self._get_explore_column_mapping()[column]
                upstream_column_refs.append(
                    ColumnRef(
                        column=explore_column.get("field", explore_column[NAME]),
                        table=explore_urn,
                    )
                )

        return upstream_column_refs

    def get_upstream_dataset_urn(self) -> List[Urn]:
        return self._get_upstream_dataset_urn()


class RegularViewUpstream(AbstractViewUpstream):
    """
    Handle the case where upstream dataset name is equal to view-name
    """

    upstream_dataset_urn: Optional[str]

    def __init__(
        self,
        view_context: LookerViewContext,
        looker_view_id_cache: LookerViewIdCache,
        config: LookMLSourceConfig,
        reporter: LookMLSourceReport,
        ctx: PipelineContext,
    ):
        super().__init__(view_context, looker_view_id_cache, config, reporter, ctx)
        self.upstream_dataset_urn = None

        self._get_upstream_dataset_urn = lru_cache(maxsize=1)(
            self.__get_upstream_dataset_urn
        )

    def __get_upstream_dataset_urn(self) -> Urn:
        # In regular case view's upstream dataset is either same as view-name or mentioned in "sql_table_name" field
        # view_context.datahub_transformed_sql_table_name() handle this condition to return dataset name
        qualified_table_name: str = _generate_fully_qualified_name(
            sql_table_name=self.view_context.datahub_transformed_sql_table_name(),
            connection_def=self.view_context.view_connection,
            reporter=self.view_context.reporter,
        )

        self.upstream_dataset_urn = make_dataset_urn_with_platform_instance(
            platform=self.view_context.view_connection.platform,
            name=qualified_table_name.lower(),
            platform_instance=self.view_context.view_connection.platform_instance,
            env=self.view_context.view_connection.platform_env or self.config.env,
        )

        return self.upstream_dataset_urn

    def get_upstream_column_ref(
        self, field_context: LookerFieldContext
    ) -> List[ColumnRef]:
        upstream_column_ref: List[ColumnRef] = []

        for column_name in field_context.column_name_in_sql_attribute():
            upstream_column_ref.append(
                ColumnRef(table=self._get_upstream_dataset_urn(), column=column_name)
            )

        return upstream_column_ref

    def get_upstream_dataset_urn(self) -> List[Urn]:
        return [self._get_upstream_dataset_urn()]


class DotSqlTableNameViewUpstream(AbstractViewUpstream):
    """
    Handle the case where upstream dataset name is mentioned as sql_table_name: ${view-name.SQL_TABLE_NAME}
    """

    upstream_dataset_urn: List[Urn]

    def __init__(
        self,
        view_context: LookerViewContext,
        looker_view_id_cache: LookerViewIdCache,
        config: LookMLSourceConfig,
        reporter: LookMLSourceReport,
        ctx: PipelineContext,
    ):
        super().__init__(view_context, looker_view_id_cache, config, reporter, ctx)
        self.upstream_dataset_urn = []

        self._get_upstream_dataset_urn = lru_cache(maxsize=1)(
            self.__get_upstream_dataset_urn
        )

    def __get_upstream_dataset_urn(self) -> List[Urn]:
        # In this case view_context.datahub_transformed_sql_table_name() refers to derived view name
        looker_view_id = get_derived_looker_view_id(
            qualified_table_name=_generate_fully_qualified_name(
                self.view_context.datahub_transformed_sql_table_name(),
                self.view_context.view_connection,
                self.view_context.reporter,
            ),
            base_folder_path=self.view_context.base_folder_path,
            looker_view_id_cache=self.looker_view_id_cache,
        )

        if looker_view_id is not None:
            self.upstream_dataset_urn = [
                looker_view_id.get_urn(
                    config=self.config,
                )
            ]

        return self.upstream_dataset_urn

    def get_upstream_column_ref(
        self, field_context: LookerFieldContext
    ) -> List[ColumnRef]:
        upstream_column_ref: List[ColumnRef] = []
        if not self._get_upstream_dataset_urn():
            return upstream_column_ref

        for column_name in field_context.column_name_in_sql_attribute():
            upstream_column_ref.append(
                ColumnRef(table=self._get_upstream_dataset_urn()[0], column=column_name)
            )

        return upstream_column_ref

    def get_upstream_dataset_urn(self) -> List[Urn]:
        return self._get_upstream_dataset_urn()


class EmptyImplementation(AbstractViewUpstream):
    def get_upstream_column_ref(
        self, field_context: LookerFieldContext
    ) -> List[ColumnRef]:
        return []

    def get_upstream_dataset_urn(self) -> List[Urn]:
        return []


def create_view_upstream(
    view_context: LookerViewContext,
    looker_view_id_cache: LookerViewIdCache,
    config: LookMLSourceConfig,
    ctx: PipelineContext,
    reporter: LookMLSourceReport,
) -> AbstractViewUpstream:

    if view_context.is_regular_case():
        return RegularViewUpstream(
            view_context=view_context,
            config=config,
            reporter=reporter,
            ctx=ctx,
            looker_view_id_cache=looker_view_id_cache,
        )

    if view_context.is_sql_table_name_referring_to_view():
        return DotSqlTableNameViewUpstream(
            view_context=view_context,
            config=config,
            reporter=reporter,
            ctx=ctx,
            looker_view_id_cache=looker_view_id_cache,
        )

    if any(
        [
            view_context.is_sql_based_derived_case(),
            view_context.is_sql_based_derived_view_without_fields_case(),
        ]
    ):

        return DerivedQueryUpstreamSource(
            view_context=view_context,
            config=config,
            reporter=reporter,
            ctx=ctx,
            looker_view_id_cache=looker_view_id_cache,
        )

    if view_context.is_direct_sql_query_case():
        return DirectQueryUpstreamSource(
            view_context=view_context,
            config=config,
            reporter=reporter,
            ctx=ctx,
            looker_view_id_cache=looker_view_id_cache,
        )

    if view_context.is_native_derived_case():
        return NativeDerivedViewUpstream(
            view_context=view_context,
            config=config,
            reporter=reporter,
            ctx=ctx,
            looker_view_id_cache=looker_view_id_cache,
        )

    reporter.report_warning(
        title="ViewUpstream Implementation Not Found",
        message="No implementation found to resolve upstream of the view",
        context=f"view_name={view_context.name()} , view_file_name={view_context.view_file_name()}",
    )

    return EmptyImplementation(
        view_context=view_context,
        config=config,
        reporter=reporter,
        ctx=ctx,
        looker_view_id_cache=looker_view_id_cache,
    )



import html
import json
import logging
from dataclasses import dataclass
from functools import lru_cache
from typing import Any, Dict, List, Optional, Tuple

from pydantic.fields import Field
from tableauserverclient import Server

import datahub.emitter.mce_builder as builder
from datahub.configuration.common import ConfigModel
from datahub.ingestion.source.tableau import tableau_constant as c
from datahub.metadata.com.linkedin.pegasus2avro.dataset import (
    DatasetLineageType,
    FineGrainedLineage,
    FineGrainedLineageDownstreamType,
    FineGrainedLineageUpstreamType,
)
from datahub.metadata.com.linkedin.pegasus2avro.schema import (
    ArrayTypeClass,
    BooleanTypeClass,
    DateTypeClass,
    NullTypeClass,
    NumberTypeClass,
    SchemaField,
    SchemaFieldDataType,
    StringTypeClass,
    TimeTypeClass,
)
from datahub.metadata.schema_classes import (
    BytesTypeClass,
    GlobalTagsClass,
    TagAssociationClass,
    UpstreamClass,
)
from datahub.sql_parsing.sqlglot_lineage import ColumnLineageInfo, SqlParsingResult

logger = logging.getLogger(__name__)


class TableauLineageOverrides(ConfigModel):
    platform_override_map: Optional[Dict[str, str]] = Field(
        default=None,
        description="A holder for platform -> platform mappings to generate correct dataset urns",
    )
    database_override_map: Optional[Dict[str, str]] = Field(
        default=None,
        description="A holder for database -> database mappings to generate correct dataset urns",
    )


class MetadataQueryException(Exception):
    pass


workbook_graphql_query = """
    {
      id
      name
      luid
      uri
      projectName
      owner {
        username
      }
      description
      uri
      createdAt
      updatedAt
      tags {
          name
      }
      sheets {
        id
      }
      dashboards {
        id
      }
      embeddedDatasources {
        id
      }
    }
"""

sheet_graphql_query = """
{
    id
    name
    path
    luid
    createdAt
    updatedAt
    tags {
        name
    }
    containedInDashboards {
        name
        path
    }
    workbook {
        id
        name
        projectName
        luid
        owner {
          username
        }
    }
    datasourceFields {
        __typename
        id
        name
        description
        datasource {
        id
        name
        }
        ... on ColumnField {
        dataCategory
        role
        dataType
        aggregation
        }
        ... on CalculatedField {
        role
        dataType
        aggregation
        formula
        }
        ... on GroupField {
        role
        dataType
        }
        ... on DatasourceField {
        remoteField {
            __typename
            id
            name
            description
            folderName
            ... on ColumnField {
            dataCategory
            role
            dataType
            aggregation
            }
            ... on CalculatedField {
            role
            dataType
            aggregation
            formula
            }
            ... on GroupField {
            role
            dataType
            }
        }
        }
    }
}
"""

dashboard_graphql_query = """
{
    id
    name
    path
    luid
    createdAt
    updatedAt
    sheets {
        id
        name
    }
    tags {
        name
    }
    workbook {
        id
        name
        projectName
        luid
        owner {
          username
        }
    }
}
"""

embedded_datasource_graphql_query = """
{
    __typename
    id
    name
    hasExtracts
    extractLastRefreshTime
    extractLastIncrementalUpdateTime
    extractLastUpdateTime
    downstreamSheets {
        name
        id
    }
    upstreamTables {
        id
        name
        database {
            name
            id
        }
        schema
        fullName
        connectionType
        description
        columnsConnection {
            totalCount
        }
    }
    fields {
        __typename
        id
        name
        description
        isHidden
        folderName
        # upstreamFields {
        #     name
        #     datasource {
        #         id
        #     }
        # }
        # upstreamColumns {
        #     name
        #     table {
        #         __typename
        #         id
        #     }
        # }
        ... on ColumnField {
            dataCategory
            role
            dataType
            defaultFormat
            aggregation
        }
        ... on CalculatedField {
            role
            dataType
            defaultFormat
            aggregation
            formula
        }
        ... on GroupField {
            role
            dataType
        }
    }
    upstreamDatasources {
        id
        name
    }
    workbook {
        id
        name
        projectName
        luid
        owner {
          username
        }
    }
}
"""

custom_sql_graphql_query = """
{
      id
      name
      query
      columns {
        id
        name
        remoteType
        description
        referencedByFields {
          datasource {
            __typename
            id
            name
            upstreamTables {
              id
              name
              database {
                name
                id
              }
              schema
              fullName
              connectionType
            }
            ... on PublishedDatasource {
              projectName
              luid
            }
            ... on EmbeddedDatasource {
              workbook {
                id
                name
                projectName
                luid
              }
            }
          }
        }
      }
      tables {
        id
        name
        database {
          name
          id
        }
        schema
        fullName
        connectionType
        description
        columnsConnection {
            totalCount
        }
      }
      connectionType
      database{
        name
        id
        connectionType
      }
}
"""


datasource_upstream_fields_graphql_query = """
{
    id
    upstreamFields {
        name
        datasource {
            id
        }
    }
    upstreamColumns {
        name
        table {
            __typename
            id
        }
    }
}
"""

published_datasource_graphql_query = """
{
    __typename
    id
    name
    luid
    hasExtracts
    extractLastRefreshTime
    extractLastIncrementalUpdateTime
    extractLastUpdateTime
    upstreamTables {
      id
      name
      database {
        name
        id
      }
      schema
      fullName
      connectionType
      description
      columnsConnection {
            totalCount
        }
    }
    fields {
        __typename
        id
        name
        description
        isHidden
        folderName
        # upstreamFields {
        #     name
        #     datasource {
        #         id
        #     }
        # }
        # upstreamColumns {
        #     name
        #     table {
        #         __typename
        #         id
        #     }
        # }
        ... on ColumnField {
            dataCategory
            role
            dataType
            defaultFormat
            aggregation
            }
        ... on CalculatedField {
            role
            dataType
            defaultFormat
            aggregation
            formula
            }
        ... on GroupField {
            role
            dataType
            }
    }
    owner {
      username
    }
    description
    uri
    projectName
    tags {
        name
    }
}
        """

database_tables_graphql_query = """
{
    id
    isEmbedded
    columns {
      remoteType
      name
    }
}
"""

database_servers_graphql_query = """
{
    name
    id
    connectionType
    extendedConnectionType
    hostName
}
"""

# https://referencesource.microsoft.com/#system.data/System/Data/OleDb/OLEDB_Enum.cs,364
FIELD_TYPE_MAPPING = {
    "INTEGER": NumberTypeClass,
    "REAL": NumberTypeClass,
    "STRING": StringTypeClass,
    "DATETIME": TimeTypeClass,
    "DATE": DateTypeClass,
    "TUPLE": ArrayTypeClass,
    "SPATIAL": NullTypeClass,
    "BOOLEAN": BooleanTypeClass,
    "TABLE": ArrayTypeClass,
    "UNKNOWN": NullTypeClass,
    "EMPTY": NullTypeClass,
    "NULL": NullTypeClass,
    "I2": NumberTypeClass,
    "I4": NumberTypeClass,
    "R4": NumberTypeClass,
    "R8": NumberTypeClass,
    "CY": NumberTypeClass,
    "BSTR": StringTypeClass,
    "IDISPATCH": NullTypeClass,
    "ERROR": NullTypeClass,
    "BOOL": BooleanTypeClass,
    "VARIANT": NullTypeClass,
    "IUNKNOWN": NullTypeClass,
    "DECIMAL": NumberTypeClass,
    "UI1": NumberTypeClass,
    "ARRAY": ArrayTypeClass,
    "BYREF": StringTypeClass,
    "I1": NumberTypeClass,
    "UI2": NumberTypeClass,
    "UI4": NumberTypeClass,
    "I8": NumberTypeClass,
    "UI8": NumberTypeClass,
    "GUID": StringTypeClass,
    "VECTOR": ArrayTypeClass,
    "FILETIME": TimeTypeClass,
    "RESERVED": NullTypeClass,
    "BYTES": BytesTypeClass,
    "STR": StringTypeClass,
    "WSTR": StringTypeClass,
    "NUMERIC": NumberTypeClass,
    "UDT": StringTypeClass,
    "DBDATE": DateTypeClass,
    "DBTIME": TimeTypeClass,
    "DBTIMESTAMP": TimeTypeClass,
    "HCHAPTER": NullTypeClass,
    "PROPVARIANT": NullTypeClass,
    "VARNUMERIC": NumberTypeClass,
    "WDC_INT": NumberTypeClass,
    "WDC_FLOAT": NumberTypeClass,
    "WDC_STRING": StringTypeClass,
    "WDC_DATETIME": TimeTypeClass,
    "WDC_BOOL": BooleanTypeClass,
    "WDC_DATE": DateTypeClass,
    "WDC_GEOMETRY": NullTypeClass,
}


def get_tags_from_params(params: List[str] = []) -> GlobalTagsClass:
    tags = [
        TagAssociationClass(tag=builder.make_tag_urn(tag.upper()))
        for tag in params
        if tag
    ]
    return GlobalTagsClass(tags=tags)


def tableau_field_to_schema_field(field, ingest_tags):
    # The check here makes sure that even if 'dataType' key exists in the 'field' dictionary but has value None,
    # it will be set as "UNKNOWN" (nativeDataType field can not be None in the SchemaField).
    # Hence, field.get("dataType", "UNKNOWN") is not enough
    nativeDataType = field.get("dataType")
    if nativeDataType is None:
        nativeDataType = "UNKNOWN"
    TypeClass = FIELD_TYPE_MAPPING.get(nativeDataType, NullTypeClass)

    schema_field = SchemaField(
        fieldPath=field["name"],
        type=SchemaFieldDataType(type=TypeClass()),
        description=make_description_from_params(
            field.get("description", ""), field.get("formula")
        ),
        nativeDataType=nativeDataType,
        globalTags=(
            get_tags_from_params(
                [
                    field.get("role", ""),
                    field.get("__typename", ""),
                    field.get("aggregation", ""),
                ]
            )
            if ingest_tags
            else None
        ),
    )

    return schema_field


@lru_cache(128)
def get_platform(connection_type: str) -> str:
    # connection_type taken from
    # https://help.tableau.com/current/api/rest_api/en-us/REST/rest_api_concepts_connectiontype.htm
    #  datahub platform mapping is found here
    # https://github.com/datahub-project/datahub/blob/master/metadata-service/war/src/main/resources/boot/data_platforms.json

    if connection_type in ("textscan", "textclean", "excel-direct", "excel", "csv"):
        platform = "external"
    elif connection_type in (
        "hadoophive",
        "hive",
        "hortonworkshadoophive",
        "maprhadoophive",
        "awshadoophive",
    ):
        platform = "hive"
    elif connection_type in ("mysql_odbc", "mysql"):
        platform = "mysql"
    elif connection_type in ("webdata-direct:oracle-eloqua", "oracle"):
        platform = "oracle"
    elif connection_type in ("tbio", "teradata"):
        platform = "teradata"
    elif connection_type in ("sqlserver"):
        platform = "mssql"
    elif connection_type in ("athena"):
        platform = "athena"
    elif connection_type.endswith("_jdbc"):
        # e.g. convert trino_jdbc -> trino
        platform = connection_type[: -len("_jdbc")]
    else:
        platform = connection_type
    return platform


@lru_cache(128)
def get_fully_qualified_table_name(
    platform: str,
    upstream_db: str,
    schema: str,
    table_name: str,
) -> str:
    if platform == "athena":
        upstream_db = ""
    database_name = f"{upstream_db}." if upstream_db else ""
    final_name = table_name.replace("[", "").replace("]", "")

    schema_name = f"{schema}." if schema else ""

    fully_qualified_table_name = f"{database_name}{schema_name}{final_name}"

    # do some final adjustments on the fully qualified table name to help them line up with source systems:
    # lowercase it
    fully_qualified_table_name = fully_qualified_table_name.lower()
    # strip double quotes, escaped double quotes and backticks
    fully_qualified_table_name = (
        fully_qualified_table_name.replace('\\"', "")
        .replace('"', "")
        .replace("\\", "")
        .replace("`", "")
    )

    if platform in ("athena", "hive", "mysql", "clickhouse"):
        # it two tier database system (athena, hive, mysql), just take final 2
        fully_qualified_table_name = ".".join(
            fully_qualified_table_name.split(".")[-2:]
        )
    else:
        # if there are more than 3 tokens, just take the final 3
        fully_qualified_table_name = ".".join(
            fully_qualified_table_name.split(".")[-3:]
        )

    return fully_qualified_table_name


@dataclass
class TableauUpstreamReference:
    database: Optional[str]
    database_id: Optional[str]
    schema: Optional[str]
    table: str

    connection_type: str

    @classmethod
    def create(
        cls, d: dict, default_schema_map: Optional[Dict[str, str]] = None
    ) -> "TableauUpstreamReference":
        # Values directly from `table` object from Tableau
        database = t_database = d.get(c.DATABASE, {}).get(c.NAME)
        database_id = d.get(c.DATABASE, {}).get(c.ID)
        schema = t_schema = d.get(c.SCHEMA)
        table = t_table = d.get(c.NAME) or ""
        t_full_name = d.get(c.FULL_NAME)
        t_connection_type = d[c.CONNECTION_TYPE]  # required to generate urn
        t_id = d[c.ID]

        parsed_full_name = cls.parse_full_name(t_full_name)
        if parsed_full_name and len(parsed_full_name) == 3:
            database, schema, table = parsed_full_name
        elif parsed_full_name and len(parsed_full_name) == 2:
            schema, table = parsed_full_name
        else:
            logger.debug(
                f"Upstream urn generation ({t_id}):"
                f"  Did not parse full name {t_full_name}: unexpected number of values",
            )

        if not schema and default_schema_map and database in default_schema_map:
            schema = default_schema_map[database]

        if database != t_database:
            logger.debug(
                f"Upstream urn generation ({t_id}):"
                f" replacing database {t_database} with {database} from full name {t_full_name}"
            )
        if schema != t_schema:
            logger.debug(
                f"Upstream urn generation ({t_id}):"
                f" replacing schema {t_schema} with {schema} from full name {t_full_name}"
            )
        if table != t_table:
            logger.debug(
                f"Upstream urn generation ({t_id}):"
                f" replacing table {t_table} with {table} from full name {t_full_name}"
            )

        # TODO: See if we can remove this -- made for redshift
        if (
            schema
            and t_table
            and t_full_name
            and t_table == t_full_name
            and schema in t_table
        ):
            logger.debug(
                f"Omitting schema for upstream table {t_id}, schema included in table name"
            )
            schema = ""

        return cls(
            database=database,
            database_id=database_id,
            schema=schema,
            table=table,
            connection_type=t_connection_type,
        )

    @staticmethod
    def parse_full_name(full_name: Optional[str]) -> Optional[List[str]]:
        # fullName is observed to be in formats:
        #  [database].[schema].[table]
        #  [schema].[table]
        #  [table]
        #  table
        #  schema

        # TODO: Validate the startswith check. Currently required for our integration tests
        if full_name is None or not full_name.startswith("["):
            return None

        return full_name.replace("[", "").replace("]", "").split(".")

    def make_dataset_urn(
        self,
        env: str,
        platform_instance_map: Optional[Dict[str, str]],
        lineage_overrides: Optional[TableauLineageOverrides] = None,
        database_hostname_to_platform_instance_map: Optional[Dict[str, str]] = None,
        database_server_hostname_map: Optional[Dict[str, str]] = None,
    ) -> str:
        (
            upstream_db,
            platform_instance,
            platform,
            original_platform,
        ) = get_overridden_info(
            connection_type=self.connection_type,
            upstream_db=self.database,
            upstream_db_id=self.database_id,
            lineage_overrides=lineage_overrides,
            platform_instance_map=platform_instance_map,
            database_hostname_to_platform_instance_map=database_hostname_to_platform_instance_map,
            database_server_hostname_map=database_server_hostname_map,
        )

        table_name = get_fully_qualified_table_name(
            original_platform,
            upstream_db or "",
            self.schema,
            self.table,
        )

        return builder.make_dataset_urn_with_platform_instance(
            platform, table_name, platform_instance, env
        )


def get_overridden_info(
    connection_type: Optional[str],
    upstream_db: Optional[str],
    upstream_db_id: Optional[str],
    platform_instance_map: Optional[Dict[str, str]],
    lineage_overrides: Optional[TableauLineageOverrides] = None,
    database_hostname_to_platform_instance_map: Optional[Dict[str, str]] = None,
    database_server_hostname_map: Optional[Dict[str, str]] = None,
) -> Tuple[Optional[str], Optional[str], str, str]:
    original_platform = platform = get_platform(connection_type)
    if (
        lineage_overrides is not None
        and lineage_overrides.platform_override_map is not None
        and original_platform in lineage_overrides.platform_override_map.keys()
    ):
        platform = lineage_overrides.platform_override_map[original_platform]

    if (
        lineage_overrides is not None
        and lineage_overrides.database_override_map is not None
        and upstream_db is not None
        and upstream_db in lineage_overrides.database_override_map.keys()
    ):
        upstream_db = lineage_overrides.database_override_map[upstream_db]

    platform_instance = (
        platform_instance_map.get(original_platform) if platform_instance_map else None
    )
    if (
        database_server_hostname_map is not None
        and upstream_db_id is not None
        and upstream_db_id in database_server_hostname_map
    ):
        hostname = database_server_hostname_map.get(upstream_db_id)
        if (
            database_hostname_to_platform_instance_map is not None
            and hostname in database_hostname_to_platform_instance_map
        ):
            platform_instance = database_hostname_to_platform_instance_map.get(hostname)

    if original_platform in (
        "athena",
        "hive",
        "mysql",
        "teradata",
    ):  # Two tier databases
        upstream_db = None

    return upstream_db, platform_instance, platform, original_platform


def make_description_from_params(description, formula):
    """
    Generate column description
    """
    final_description = ""
    if description:
        final_description += f"{description}\n\n"
    if formula:
        final_description += f"formula: {formula}"
    return final_description


def make_upstream_class(
    parsed_result: Optional[SqlParsingResult],
) -> List[UpstreamClass]:
    upstream_tables: List[UpstreamClass] = []

    if parsed_result is None:
        return upstream_tables

    for dataset_urn in parsed_result.in_tables:
        upstream_tables.append(
            UpstreamClass(type=DatasetLineageType.TRANSFORMED, dataset=dataset_urn)
        )
    return upstream_tables


def make_fine_grained_lineage_class(
    parsed_result: Optional[SqlParsingResult],
    dataset_urn: str,
    out_columns: List[Dict[Any, Any]],
) -> List[FineGrainedLineage]:
    # 1) fine grained lineage links are case sensitive
    # 2) parsed out columns are always lower cased
    # 3) corresponding Custom SQL output columns can be in any case lower/upper/mix
    #
    # we need a map between 2 and 3 that will be used during building column level linage links (see below)
    out_columns_map = {
        col.get(c.NAME, "").lower(): col.get(c.NAME, "") for col in out_columns
    }

    fine_grained_lineages: List[FineGrainedLineage] = []

    if parsed_result is None:
        return fine_grained_lineages

    cll: List[ColumnLineageInfo] = (
        parsed_result.column_lineage if parsed_result.column_lineage is not None else []
    )

    for cll_info in cll:
        downstream = (
            [
                builder.make_schema_field_urn(
                    dataset_urn,
                    out_columns_map.get(
                        cll_info.downstream.column.lower(),
                        cll_info.downstream.column,
                    ),
                )
            ]
            if cll_info.downstream is not None
            and cll_info.downstream.column is not None
            else []
        )
        upstreams = [
            builder.make_schema_field_urn(column_ref.table, column_ref.column)
            for column_ref in cll_info.upstreams
        ]

        fine_grained_lineages.append(
            FineGrainedLineage(
                downstreamType=FineGrainedLineageDownstreamType.FIELD,
                downstreams=downstream,
                upstreamType=FineGrainedLineageUpstreamType.FIELD_SET,
                upstreams=upstreams,
            )
        )

    return fine_grained_lineages


def get_unique_custom_sql(custom_sql_list: List[dict]) -> List[dict]:
    unique_custom_sql = []
    for custom_sql in custom_sql_list:
        unique_csql = {
            "id": custom_sql.get("id"),
            "name": custom_sql.get("name"),
            # We assume that this is unsupported custom sql if "actual tables that this query references"
            # are missing from api result.
            "isUnsupportedCustomSql": True if not custom_sql.get("tables") else False,
            "query": custom_sql.get("query"),
            "connectionType": custom_sql.get("connectionType"),
            "columns": custom_sql.get("columns"),
            "tables": custom_sql.get("tables"),
            "database": custom_sql.get("database"),
        }
        datasource_for_csql = []
        for column in custom_sql.get("columns", []):
            for field in column.get("referencedByFields", []):
                datasource = field.get("datasource")
                if datasource not in datasource_for_csql and datasource is not None:
                    datasource_for_csql.append(datasource)

        unique_csql["datasources"] = datasource_for_csql
        unique_custom_sql.append(unique_csql)
    return unique_custom_sql


def clean_query(query: str) -> str:
    """
    Clean special chars in query
    """
    query = query.replace("<<", "<").replace(">>", ">").replace("\n\n", "\n")
    query = html.unescape(query)
    return query


def make_filter(filter_dict: dict) -> str:
    filter = ""
    for key, value in filter_dict.items():
        if filter:
            filter += ", "
        filter += f"{key}: {json.dumps(value)}"
    return filter


def query_metadata_cursor_based_pagination(
    server: Server,
    main_query: str,
    connection_name: str,
    first: int,
    after: Optional[str],
    qry_filter: str = "",
) -> dict:
    query = f"""
        query GetItems(
          $first: Int,
          $after: String
        ) {{
            {connection_name} (  first: $first, after: $after, filter:{{ {qry_filter} }})
            {{
                nodes {main_query}
                pageInfo {{
                    hasNextPage
                    endCursor
                }}
            }}
    }}"""  # {{ is to escape { character of f-string

    result = server.metadata.query(
        query=query,
        variables={
            "first": first,
            "after": after,
        },
    )

    return result


def get_filter_pages(query_filter: dict, page_size: int) -> List[dict]:
    filter_pages = [query_filter]
    # If this is primary id filter, so we can use divide this query list into
    # multiple requests each with smaller filter list (of order page_size).
    # It is observed in the past that if a list of primary ids grows beyond
    # a few ten thousand, then tableau server responds with empty response
    # causing below error:
    # tableauserverclient.server.endpoint.exceptions.NonXMLResponseError: b''
    if (
        len(query_filter.keys()) == 1
        and query_filter.get(c.ID_WITH_IN)
        and isinstance(query_filter[c.ID_WITH_IN], list)
        and len(query_filter[c.ID_WITH_IN]) > 100 * page_size
    ):
        ids = query_filter[c.ID_WITH_IN]
        filter_pages = [
            {
                c.ID_WITH_IN: ids[
                    start : (
                        start + page_size if start + page_size < len(ids) else len(ids)
                    )
                ]
            }
            for start in range(0, len(ids), page_size)
        ]

    return filter_pages



# mypy: ignore-errors
# flake8: noqa

# This file is autogenerated by /metadata-ingestion/scripts/avro_codegen.py
# Do not modify manually!

# pylint: skip-file
# fmt: off
# isort: skip_file
from .....schema_classes import AccessClass
from .....schema_classes import AccessLevelClass
from .....schema_classes import AuditStampClass
from .....schema_classes import BrowsePathEntryClass
from .....schema_classes import BrowsePathsClass
from .....schema_classes import BrowsePathsV2Class
from .....schema_classes import ChangeAuditStampsClass
from .....schema_classes import CostClass
from .....schema_classes import CostCostClass
from .....schema_classes import CostCostDiscriminatorClass
from .....schema_classes import CostTypeClass
from .....schema_classes import DataPlatformInstanceClass
from .....schema_classes import DeprecationClass
from .....schema_classes import DocumentationClass
from .....schema_classes import DocumentationAssociationClass
from .....schema_classes import EdgeClass
from .....schema_classes import EmbedClass
from .....schema_classes import FabricTypeClass
from .....schema_classes import FieldFormPromptAssociationClass
from .....schema_classes import FormAssociationClass
from .....schema_classes import FormPromptAssociationClass
from .....schema_classes import FormPromptFieldAssociationsClass
from .....schema_classes import FormVerificationAssociationClass
from .....schema_classes import FormsClass
from .....schema_classes import GlobalTagsClass
from .....schema_classes import GlossaryTermAssociationClass
from .....schema_classes import GlossaryTermsClass
from .....schema_classes import IncidentSummaryDetailsClass
from .....schema_classes import IncidentsSummaryClass
from .....schema_classes import InputFieldClass
from .....schema_classes import InputFieldsClass
from .....schema_classes import InstitutionalMemoryClass
from .....schema_classes import InstitutionalMemoryMetadataClass
from .....schema_classes import MLFeatureDataTypeClass
from .....schema_classes import MediaClass
from .....schema_classes import MediaTypeClass
from .....schema_classes import MetadataAttributionClass
from .....schema_classes import OperationClass
from .....schema_classes import OperationSourceTypeClass
from .....schema_classes import OperationTypeClass
from .....schema_classes import OriginClass
from .....schema_classes import OriginTypeClass
from .....schema_classes import OwnerClass
from .....schema_classes import OwnershipClass
from .....schema_classes import OwnershipSourceClass
from .....schema_classes import OwnershipSourceTypeClass
from .....schema_classes import OwnershipTypeClass
from .....schema_classes import RoleAssociationClass
from .....schema_classes import SiblingsClass
from .....schema_classes import StatusClass
from .....schema_classes import SubTypesClass
from .....schema_classes import TagAssociationClass
from .....schema_classes import TimeStampClass
from .....schema_classes import VersionTagClass
from .....schema_classes import WindowDurationClass


Access = AccessClass
AccessLevel = AccessLevelClass
AuditStamp = AuditStampClass
BrowsePathEntry = BrowsePathEntryClass
BrowsePaths = BrowsePathsClass
BrowsePathsV2 = BrowsePathsV2Class
ChangeAuditStamps = ChangeAuditStampsClass
Cost = CostClass
CostCost = CostCostClass
CostCostDiscriminator = CostCostDiscriminatorClass
CostType = CostTypeClass
DataPlatformInstance = DataPlatformInstanceClass
Deprecation = DeprecationClass
Documentation = DocumentationClass
DocumentationAssociation = DocumentationAssociationClass
Edge = EdgeClass
Embed = EmbedClass
FabricType = FabricTypeClass
FieldFormPromptAssociation = FieldFormPromptAssociationClass
FormAssociation = FormAssociationClass
FormPromptAssociation = FormPromptAssociationClass
FormPromptFieldAssociations = FormPromptFieldAssociationsClass
FormVerificationAssociation = FormVerificationAssociationClass
Forms = FormsClass
GlobalTags = GlobalTagsClass
GlossaryTermAssociation = GlossaryTermAssociationClass
GlossaryTerms = GlossaryTermsClass
IncidentSummaryDetails = IncidentSummaryDetailsClass
IncidentsSummary = IncidentsSummaryClass
InputField = InputFieldClass
InputFields = InputFieldsClass
InstitutionalMemory = InstitutionalMemoryClass
InstitutionalMemoryMetadata = InstitutionalMemoryMetadataClass
MLFeatureDataType = MLFeatureDataTypeClass
Media = MediaClass
MediaType = MediaTypeClass
MetadataAttribution = MetadataAttributionClass
Operation = OperationClass
OperationSourceType = OperationSourceTypeClass
OperationType = OperationTypeClass
Origin = OriginClass
OriginType = OriginTypeClass
Owner = OwnerClass
Ownership = OwnershipClass
OwnershipSource = OwnershipSourceClass
OwnershipSourceType = OwnershipSourceTypeClass
OwnershipType = OwnershipTypeClass
RoleAssociation = RoleAssociationClass
Siblings = SiblingsClass
Status = StatusClass
SubTypes = SubTypesClass
TagAssociation = TagAssociationClass
TimeStamp = TimeStampClass
VersionTag = VersionTagClass
WindowDuration = WindowDurationClass

# fmt: on



import datetime
import logging
import re
import time
from dataclasses import dataclass
from typing import Dict, Iterable

from pydantic import Field

from datahub.configuration.common import ConfigModel, OperationalError
from datahub.ingestion.api.common import PipelineContext
from datahub.ingestion.api.decorators import (
    SupportStatus,
    config_class,
    platform_name,
    support_status,
)
from datahub.ingestion.api.source import Source, SourceReport
from datahub.ingestion.api.workunit import MetadataWorkUnit

logger = logging.getLogger(__name__)


class DataHubGcSourceConfig(ConfigModel):
    cleanup_expired_tokens: bool = Field(
        default=True,
        description="Whether to clean up expired tokens or not",
    )
    truncate_indices: bool = Field(
        default=True,
        description="Whether to truncate elasticsearch indices or not which can be safely truncated",
    )
    truncate_index_older_than_days: int = Field(
        default=30,
        description="Indices older than this number of days will be truncated",
    )
    truncation_watch_until: int = Field(
        default=10000,
        description="Wait for truncation of indices until this number of documents are left",
    )
    truncation_sleep_between_seconds: int = Field(
        default=30,
        description="Sleep between truncation monitoring.",
    )


@dataclass
class DataHubGcSourceReport(SourceReport):
    expired_tokens_revoked: int = 0


@platform_name("DataHubGc")
@config_class(DataHubGcSourceConfig)
@support_status(SupportStatus.TESTING)
class DataHubGcSource(Source):
    def __init__(self, ctx: PipelineContext, config: DataHubGcSourceConfig):
        self.ctx = ctx
        self.config = config
        self.report = DataHubGcSourceReport()
        self.graph = ctx.require_graph("The DataHubGc source")

    @classmethod
    def create(cls, config_dict, ctx):
        config = DataHubGcSourceConfig.parse_obj(config_dict)
        return cls(ctx, config)

    def get_workunits_internal(
        self,
    ) -> Iterable[MetadataWorkUnit]:
        if self.config.cleanup_expired_tokens:
            self.revoke_expired_tokens()
        if self.config.truncate_indices:
            self.truncate_indices()
        yield from []

    def truncate_indices(self) -> None:
        self._truncate_timeseries_helper(aspect_name="operation", entity_type="dataset")
        self._truncate_timeseries_helper(
            aspect_name="datasetusagestatistics", entity_type="dataset"
        )
        self._truncate_timeseries_helper(
            aspect_name="chartUsageStatistics", entity_type="chart"
        )
        self._truncate_timeseries_helper(
            aspect_name="dashboardUsageStatistics", entity_type="dashboard"
        )

    def _truncate_timeseries_helper(self, aspect_name: str, entity_type: str) -> None:
        self._truncate_timeseries_with_watch_optional(
            aspect_name=aspect_name, entity_type=entity_type, watch=False
        )
        self._truncate_timeseries_with_watch_optional(
            aspect_name=aspect_name, entity_type=entity_type, watch=True
        )

    def _truncate_timeseries_with_watch_optional(
        self, aspect_name: str, entity_type: str, watch: bool
    ) -> None:
        graph = self.graph
        assert graph is not None
        if watch:
            to_delete = 1
            while to_delete > 0:
                response = self.truncate_timeseries_util(
                    aspect=aspect_name,
                    dry_run=watch,
                    days_ago=self.config.truncate_index_older_than_days,
                    entity_type=entity_type,
                )
                val = response.get("value", "")
                if "This was a dry run" not in val or "out of" not in val:
                    return
                prev_to_delete = to_delete
                to_delete, total = re.findall(r"\d+", val)[:2]
                to_delete = int(to_delete)
                if to_delete <= 0:
                    logger.info("Nothing to delete.")
                    return
                logger.info(f"to_delete {to_delete} / {total}")
                if to_delete == prev_to_delete:
                    logger.info("Seems to be stuck. Ending the loop.")
                    break
                elif to_delete < self.config.truncation_watch_until:
                    logger.info("Too small truncation. Not going to watch.")
                    return
                else:
                    time.sleep(self.config.truncation_sleep_between_seconds)
        else:
            self.truncate_timeseries_util(
                aspect=aspect_name,
                dry_run=watch,
                days_ago=self.config.truncate_index_older_than_days,
                entity_type=entity_type,
            )

    def x_days_ago_millis(self, days: int) -> int:
        x_days_ago_datetime = datetime.datetime.now(
            datetime.timezone.utc
        ) - datetime.timedelta(days=days)
        return int(x_days_ago_datetime.timestamp() * 1000)

    def truncate_timeseries_util(
        self,
        aspect: str,
        days_ago: int,
        dry_run: bool = True,
        entity_type: str = "dataset",
    ) -> Dict:
        graph = self.graph
        assert graph is not None

        gms_url = graph._gms_server
        if not dry_run:
            logger.info(
                f"Going to truncate timeseries for {aspect} for {gms_url} older than {days_ago} days"
            )
        days_ago_millis = self.x_days_ago_millis(days_ago)
        url = f"{gms_url}/operations?action=truncateTimeseriesAspect"
        try:
            response = graph._post_generic(
                url=url,
                payload_dict={
                    "entityType": entity_type,
                    "aspect": aspect,
                    "endTimeMillis": days_ago_millis,
                    "dryRun": dry_run,
                },
            )
            # logger.info(f"Response: {response}")
        except OperationalError:
            response = graph._post_generic(
                url=url,
                payload_dict={
                    "entityType": entity_type,
                    "aspect": aspect,
                    "endTimeMillis": days_ago_millis,
                    "dryRun": dry_run,
                    "forceDeleteByQuery": True,
                },
            )
            # logger.info(f"Response: {response}")
        return response

    def revoke_expired_tokens(self) -> None:
        total = 1
        while total > 0:
            expired_tokens_res = self._get_expired_tokens()
            list_access_tokens = expired_tokens_res.get("listAccessTokens", {})
            tokens = list_access_tokens.get("tokens", [])
            total = list_access_tokens.get("total", 0)
            for token in tokens:
                self.report.expired_tokens_revoked += 1
                token_id = token["id"]
                self._revoke_token(token_id)

    def _revoke_token(self, token_id: str) -> None:
        assert self.graph is not None
        self.graph.execute_graphql(
            query="""mutation revokeAccessToken($tokenId: String!) {
  revokeAccessToken(tokenId: $tokenId)
}
""",
            variables={"tokenId": token_id},
        )

    def _get_expired_tokens(self) -> dict:
        assert self.graph is not None
        return self.graph.execute_graphql(
            query="""query listAccessTokens($input: ListAccessTokenInput!) {
  listAccessTokens(input: $input) {
    start
    count
    total
    tokens {
      urn
      type
      id
      name
      description
      actorUrn
      ownerUrn
      createdAt
      expiresAt
      __typename
    }
    __typename
  }
}
""",
            variables={
                "input": {
                    "start": 0,
                    "count": 10,
                    "filters": [
                        {
                            "field": "expiresAt",
                            "values": [str(int(time.time() * 1000))],
                            "condition": "LESS_THAN",
                        }
                    ],
                }
            },
        )

    def get_report(self) -> SourceReport:
        return self.report



import logging

from gql.transport.requests import log as requests_logger

from datahub.api.circuit_breaker.assertion_circuit_breaker import (
    AssertionCircuitBreaker,
    AssertionCircuitBreakerConfig,
)
from datahub.api.circuit_breaker.operation_circuit_breaker import (
    OperationCircuitBreaker,
    OperationCircuitBreakerConfig,
)

requests_logger.setLevel(logging.WARNING)
